"""
   Descp: This class is used to build a GraphQL query.

   Created on: 26-feb-2020

   Copyright 2020-2021 Youssef 'FRYoussef' El Faqir El Rhazoui 
        <f.r.youssef@hotmail.com>
"""

from typing import List

from dao_analyzer.web.apps.common.api.graphql.query import Query

class QueryBuilder():

    def __init__(self, queries: List[Query] = None):
        self.queries: List[Query] = queries if queries else list()


    def add_query(self, query: Query):
        self.queries.append(query)


    def build(self) -> str:
        query: str = '{ '
        
        for q in self.queries:
            q_filter: str = ''
            for f in q.filters:
                q_filter += f'{f}: {q.filters[f]}, '

            q_body: str = ''
            if type(q.body) == list:
                q_body += '{ '
                for attr in q.body:
                    q_body += f'{attr} '
                q_body += '}'

            elif type(q.body) == Query:
                q_builder: QueryBuilder = QueryBuilder([q.body])
                q_body += q_builder.build()

            query += f'{q.header}({q_filter}){q_body} '

        query += '}'
        return query


"""
   Descp: Controller class for all the charts.

   Created on: 10-sep-2020

   Copyright 2020-2021 Youssef 'FRYoussef' El Faqir El Rhazoui
        <f.r.youssef@hotmail.com>
"""

import logging
from dash import html
from dash.dependencies import Input, Output, State
from dao_analyzer.web.apps.common.business.transfers.organization.organization_list import OrganizationList

from dao_analyzer.web.apps.common.presentation.charts.layout.chart_pane_layout \
    import ChartPaneLayout
from dao_analyzer.web.apps.common.business.i_metric_adapter import IMetricAdapter

class ChartController():

    def __init__(self, css_id: str, layout: ChartPaneLayout, adapter: IMetricAdapter) -> None:
        self._layout: ChartPaneLayout = layout
        self._adapter: IMetricAdapter = adapter
        self._css_id: str = css_id


    @property
    def layout(self) -> ChartPaneLayout:
        return self._layout


    def bind_callback(self, app) -> None:

        @app.callback(
             Output(self._css_id, 'children'),
            [Input('org-dropdown', 'value')],
            State('org-dropdown', 'options'),
        )
        def update_chart(org_id, org_options):
            if not org_id:
                self._layout.get_layout()

            try:
                data = self._adapter.get_plot_data(org_id, OrganizationList.from_dict_representation(org_options))
            except ValueError as e:
                logging.exception(e, exc_info=True)
                return html.Span("Error showing graph. See logs for details.", className='text-danger')
            else:
                return self._layout.fill_child(plot_data=data)



"""
   Descp: The interface which each figure must implement.

   Created on: 09-sep-2020

   Copyright 2020-2021 Youssef 'FRYoussef' El Faqir El Rhazoui
        <f.r.youssef@hotmail.com>
"""

import abc
from typing import Dict

from dao_analyzer.web.apps.common.presentation.charts.layout.figure.figure_configuration \
    import FigureConfiguration

class Figure(metaclass=abc.ABCMeta):


    def __init__(self) -> None:
        self.__configuration: FigureConfiguration = FigureConfiguration()


    @classmethod
    def __subclasshook__(cls, subclass):
        return (hasattr(subclass, 'get_empty_figure') and
                callable(subclass.get_empty_figure) and
                hasattr(subclass, 'get_figure') and
                callable(subclass.get_figure) and
                hasattr(subclass, 'get_empty_plot_data') and
                callable(subclass.get_empty_plot_data) or
                NotImplemented)

    
    @property
    def configuration(self) -> FigureConfiguration:
        return self.__configuration


    @abc.abstractmethod
    def get_empty_figure(self) -> Dict:
        """
        Returns an empty figure. 
        """
        raise NotImplementedError


    @abc.abstractmethod
    def get_figure(self, plot_data: Dict) -> Dict:
        """
        Returns the figure filled with the plot_data argument.
        """
        raise NotImplementedError


    @abc.abstractmethod
    def get_empty_plot_data(self) -> Dict:
        """
        Returns an empty dictionary with the plot_data schema.
        """
        raise NotImplementedError



import pytest

from dash.testing.application_runners import import_app

@pytest.mark.skip("Not Implemented yet")
def test_render_component(dash_duo):
    raise NotImplementedError

    app = import_app('dao-analyzer')

    dash_duo.start_server(app)



import abc

from dash import html

class ILayout(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def get_layout(self) -> html.Div: 
        raise NotImplementedError

    @classmethod
    def pane_id(cls) -> int:
        pane_id: int = cls.PANE_ID
        cls.PANE_ID += 1
        return pane_id


"""
   Descp: This file is used to store color values and functions related to.

   Created on: 30-mar-2020

   Copyright 2020-2021 Youssef 'FRYoussef' El Faqir El Rhazoui
        <f.r.youssef@hotmail.com>
"""

from typing import Tuple


DARK_BLUE = '#2471a3'
LIGHT_BLUE = '#d4e6f1'
DARK_RED = '#F44336'
LIGHT_RED = '#EF9A9A'
DARK_GREEN = '#4CAF50'
LIGHT_GREEN = '#A5D6A7'
DARK_PURPLE = '#9C27B0'
LIGHT_PURPLE = '#CE93D8'
LIGHT_YELLOW = '#FFF59D'
DARK_YELLOW = '#FFEB3B'
LIGHT_ORANGE = '#FFCC80'
DARK_ORANGE = '#FF9800'
BLACK = 'black'
WHITE = 'white'
GRID_COLOR = '#B0BEC5'
TICKFONT_COLOR = '#696969'


def hex_to_rgba(color: str, alpha: float = 1.0) -> Tuple:
    color = color.lstrip('#')
    color = tuple(int(color[i:i+2], 16) for i in (0, 2, 4))
    color = color + (alpha,)

    return color



"""
   Descp: Strategy pattern to create casted votes rate.

   Created on: 05-nov-2020

   Copyright 2020-2021 Youssef 'FRYoussef' El Faqir El Rhazoui
        <f.r.youssef@hotmail.com>
"""
import pandas as pd
from typing import List

from dao_analyzer.web.apps.common.data_access.daos.metric.strategy import IMetricStrategy
from dao_analyzer.web.apps.common.business.transfers.stacked_serie import StackedSerie
from dao_analyzer.web.apps.common.business.transfers.serie import Serie 
import dao_analyzer.web.apps.common.data_access.pandas_utils as pd_utl
from dao_analyzer.web.apps.aragon.data_access.daos.metric.strategy.st_cast_type \
    import StCastType


class StCastedVotesRate(IMetricStrategy):

    CAST_VOTE_FOR: int = 0
    CAST_VOTE_AGAINST: int = 1


    def __init__(self, m_type) -> None:
        self.__type = m_type


    def clean_df(self, df: pd.DataFrame) -> pd.DataFrame:
        pass


    def process_data(self, df: pd.DataFrame) -> StackedSerie:
        if pd_utl.is_an_empty_df(df):
            return StackedSerie()
        
        casts: StackedSerie = StCastType().process_data(df)

        rate: List[float] = self.__calculate_rate(m_casts=casts)

        metric: StackedSerie = StackedSerie(
            serie = Serie(x=casts.get_serie()), 
            y_stack = [rate])

        return metric


    def __calculate_rate(self, m_casts: StackedSerie) -> List[float]:
        rates: List[float] = []
        casts_for: List[int] = m_casts.get_i_stack(1)
        casts_against: List[int] = m_casts.get_i_stack(0)
        casts_type: List[int] = casts_against \
            if self.__type == self.CAST_VOTE_AGAINST else casts_for

        for i, _ in enumerate(casts_for):
            denominator: int = casts_for[i] + casts_against[i]
            rate: float = casts_type[i] / denominator if denominator else None
            rates.append(rate)

        return rates



import multiprocessing

wsgi_app = 'dao_analyzer.web:create_app()'
bind = ':80'
timeout = 1200
proc_name = 'DAO-Analyzer'
loglevel = 'info'
errorlog = '-'
workers = multiprocessing.cpu_count() * 2 + 1
raw_env = [
  # A bug with gunicorn and dash makes it so its always hot-reloading
  "DASH_HOT_RELOAD=false",
]



"""
    Descp: Metadata definitions

    Created on: 23-nov-2021

    Copyright 2021 David Davó
        <david@ddavo.me>
"""
from json.encoder import JSONEncoder
from typing import Dict
import json
from functools import total_ordering
from datetime import datetime, timezone

from . import config

@total_ordering
class Block:
    def __init__(self, init=None):
        self.number = 0
        self.id = None
        self.timestamp = datetime.min

        if isinstance(init, dict):
            self.number = int(init["number"]) if "number" in init else self.number
            self.id = init["id"] if "id" in init else self.id

            if "timestamp" in init:
                if init["timestamp"].isdigit():
                    self.timestamp = datetime.fromtimestamp(int(init["timestamp"]))
                else:
                    self.timestamp = datetime.fromisoformat(init["timestamp"])
            
        if self.timestamp.tzinfo is None:
            self.timestamp = self.timestamp.replace(tzinfo=timezone.utc)

    def __eq__(self, other):
        if isinstance(other, Block):
            # Both shouldn't be null
            return not self.id and not other.id and self.id == other.id
        else:
            return False

    def __lt__(self, other):
        return self.number and self.number < other.number

    def toDict(self):
        return {
            "number": self.number,
            "id": self.id,
            "timestamp": self.timestamp.isoformat()
        }

    def __str__(self):
        return self.toDict().__str__()

class CollectorMetaData:
    def __init__(self, c: str, d = None):
        self.block = Block()
        self._collector: str = c
        self.last_update: datetime = datetime.now(timezone.utc)

        if d:
            self.block = Block(d["block"]) if "block" in d else None
            self.last_update = datetime.fromisoformat(d["last_update"])

        if self.last_update.tzinfo is None:
            self.last_update = self.last_update.replace(tzinfo=timezone.utc)

    def toDict(self):
        return {
            "block": self.block,
            "last_update": self.last_update.isoformat()
        }

    def __eq__(self, other):
        if isinstance(other, CollectorMetaData):
            return self._collector == other._collector and self.block == other.block
        else:
            return False

class MetadataEncoder(JSONEncoder):
    def default(self, o):
        if hasattr(o, 'toDict'):
            return o.toDict()
        else:
            return super().default(o)

class RunnerMetadata:
    def __init__(self, runner):
        self._path = runner.basedir / 'metadata.json'
        self.collectorMetaData: Dict[str, CollectorMetaData] = {}
        self.errors: Dict[str, str] = {}
        self._setPrev()

    def _setPrev(self):
        self._prev = (self.errors.copy(), self.collectorMetaData.copy())
    
    @property
    def dirty(self) -> bool:
        return (self.errors, self.collectorMetaData) != self._prev

    def __getitem__(self, key):
        if key not in self.collectorMetaData:
            self.collectorMetaData[key] = CollectorMetaData(key)
        return self.collectorMetaData[key]
    
    def __setitem__(self, key, val):
        self.collectorMetaData[key] = val
        self.ifdump()

    def __delitem__(self, key):
        del self.collectorMetaData[key]

    def load(self):
        with open(self._path, 'r') as f:
            j = json.load(f)
            self.collectorMetaData = {k:CollectorMetaData(k,v) for k,v in j["metadata"].items()}
            self.errors = j["errors"]
            self._setPrev()

    def dump(self):
        with open(self._path, 'w') as f:
            json.dump({
                "metadata": self.collectorMetaData,
                "errors": self.errors
            }, f, 
                indent=2 if config.debug else None, 
                cls=MetadataEncoder)

    def ifdump(self):
        if self.dirty:
            self.dump()

    def __enter__(self):
        if self._path.is_file():
            self.load()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.dump()



import pandas as pd
import requests
import logging
from functools import partial
from tqdm import tqdm
from time import sleep
from typing import Union

import numpy as np

from ..metadata import Block
from .. import config

from . import ENDPOINTS
from .common import NetworkCollector, solve_decimals
from .cryptocompare import cc_postprocessor
from .graphql import GraphQLCollector

MSG_NO_TOKENS_FOUND = "No tokens found"

class BlockscoutBallancesCollector(NetworkCollector):
    ERR_SLEEP = 60

    def __init__(self, runner, base: GraphQLCollector, name: str='tokenBalances', network: str='mainnet', addr_key: str='id'):
        """ Initializes a ballance collector that uses blockscout
        
        Parameters:
            name (str): The name of the collector (and filename)
            runner (Runner): The runner in which it's based
            network (str): The network to run on
            base (GraphQLCollector): The DAOs/organizations collector to get the DAO list from
            index_key (str): The key to use to access the address in the base array
        """
        super().__init__(name, runner, network)
        self.base = base
        self.addr_key = addr_key

    def verify(self) -> bool:
        if config.skip_token_balances:
            logging.warning('Skipping token balances because --skip-token-balances flag was set')
            return False
        else:
            return super().verify()

    @property
    def endpoint(self) -> str:
        return ENDPOINTS[self.network]['blockscout']

    def _get_from_address(self, addr: str, retry: int = 0, maxretries: int = 3, block: Union[int, Block] = None, ignore_errors=False) -> pd.DataFrame: # noqa: C901
        if retry >= maxretries:
            raise ValueError(f"Too many retries {retry}/{maxretries}")

        blockn = block
        if blockn is None:
            blockn = 'latest'
        elif isinstance(blockn, Block):
            blockn = blockn.number

        r = requests.get(ENDPOINTS[self.network]['blockscout'], params={
            'module': 'account',
            'action': 'tokenlist',
            'block': blockn,
            'address': addr
        })

        if (r.ok):
            j = r.json()
            if j['status'] == str(1) and j['message'] == 'OK':
                df = pd.DataFrame(j['result'])
                # Only ERC-20 tokens (not NFTs or others)
                df = df[df.type == 'ERC-20']
                df = df.replace(r'^\s*$', np.nan, regex=True)
                df = df.dropna()

                # Calculate decimals
                solve_decimals(df)

                # Add index
                df['address'] = addr
                df['network'] = self.network
                df['id'] = 'token-' + df['contractAddress'] + '-org-' + df['address']
                return df
            elif j['message'] == MSG_NO_TOKENS_FOUND:
                return pd.DataFrame()
            else:
                logging.warning(f"Status {j['status']}, message: {j['message']}")
                return pd.DataFrame()
        elif r.status_code == 429: # Too many requests
            logging.warning(f"Too many requests, sleep and retry {retry}/{maxretries} time")
            sleep(self.ERR_SLEEP)
            return self._get_from_address(addr, retry=retry+1, maxretries=maxretries)
        elif r.status_code == 503:
            logging.warning(f"Service unavailable, sleep and retry {retry}/{maxretries} time")
            sleep(self.ERR_SLEEP)
            return self._get_from_address(addr, retry=retry+1, maxretries=maxretries)
        elif r.status_code == 504: # Gateway Time-out (Response too large)
            logging.warning(f"Requests returned Gateway Time-out, ignoring response for addr {addr}")
            return pd.DataFrame() 
        else:
            logging.error(f'Requests failed for address "{addr}" with status code {r.status_code}: {r.reason}')
            if ignore_errors:
                return pd.DataFrame()
            else:
                raise ValueError(f"Requests failed for address {addr[:12]}... with status code {r.status_code}: {r.reason}")

    def run(self, force=False, block: Block = None, prev_block: Block = None, **kwargs):
        # For each of the DAOs in the df, get the token balance
        addresses = self.base.df[self.addr_key].drop_duplicates()

        if addresses.empty:
            logging.warning("No addresses returned, not running blockscout collector")
            return

        ptqdm = partial(tqdm, delay=1, desc="Requesting token balances", 
            unit='req', dynamic_ncols=True)
        toApply = partial(self._get_from_address, block=block, ignore_errors=True)
        df = pd.concat(map(toApply, ptqdm(addresses)), ignore_index=True)

        df = cc_postprocessor(df)
        df = df.rename(columns={
            'name': 'tokenName',
            # Replace only if addr_key is not 'id'
            'address': self.addr_key if self.addr_key != 'id' else 'address'
        })
        
        self._update_data(df, force)


"""
   Descp: Interface for requesters.

   Created on: 16-jul-2020

   Copyright 2020-2021 Youssef 'FRYoussef' El Faqir El Rhazoui
        <f.r.youssef@hotmail.com>
"""

import abc
from typing import Any

class IRequester(metaclass=abc.ABCMeta):
    @classmethod
    def __subclasshook__(cls, subclass):
        return (hasattr(subclass, 'request') and
            callable(subclass.request) or
            NotImplemented)


    @abc.abstractmethod
    def request(self, *args) -> Any:
        """
        Common function to request data.
        Arguments: 
                * args: a list of arguments to request
        Return:
            The return value is tied to the interface implementation. 
        """
        raise NotImplementedError



"""
   Descp: Active token holders test.

   Created on: 22-oct-2020

   Copyright 2020-2021 Youssef 'FRYoussef' El Faqir El Rhazoui
        <f.r.youssef@hotmail.com>
"""

import unittest
from typing import List
import pandas as pd

from test.mocks.unix_date_builder import UnixDateBuilder
from dao_analyzer.web.apps.aragon.data_access.daos.metric.strategy.\
    st_active_token_holders import StActiveTokenHolders

from dao_analyzer.web.apps.common.business.transfers.stacked_serie import StackedSerie


class StActiveTokenHoldersTest(unittest.TestCase):

    def __check_lists(self, df: pd.DataFrame, out: List[int]) -> None:
        strategy: StActiveTokenHolders = StActiveTokenHolders()
        result: StackedSerie = strategy.process_data(df=df)

        self.assertListEqual(out, result.get_i_stack(i_stack=0))


    def test_process_data(self):
        bl: UnixDateBuilder = UnixDateBuilder()
        bl.sub(month=3).change(day=27, hour=0, minute=0, second=0)

        in_df: pd.DataFrame = pd.DataFrame([
            {'voter': 1, 'trash': 'trash', 'createdAt': bl.unix()},#today_year-(today_month-3)-27T00:00:00+00:00
            {'voter': 1, 'trash': 'trash', 'createdAt': bl.change(hour=5).unix()},#today_year-(today_month-3)-27T05:00:00+00:00
            {'entity': 1, 'trash': 'trash', 'date': bl.add(month=2).unix()},#today_year-(today_month-1)-27T05:00:00+00:00
            {'entity': 2, 'trash': 'trash', 'date': bl.change(day=3, hour=23).unix()},#today_year-(today_month-1)-03T23:00:00+00:00
            {'originalCreator': 2, 'trash': 'trash', 'startDate': bl.add(month=1).change(minute=59, second=59).unix()},#today_year-today_month-03T23:59:59+00:00
            {'entity': 3, 'trash': 'trash', 'date': bl.unix()},#today_year-today_month-03T23:59:59+00:00
            {'originalCreator': 2, 'trash': 'trash', 'startDate': bl.add(second=1).change(day=7, hour=17).unix()},#today_year-today_month-07T17:00:00+00:00
        ])
        out: List[int] = [1, 0, 2, 2]

        self.__check_lists(df=in_df, out=out)


if __name__ == "__main__":
    unittest.main()



from typing import Any

import pandas as pd
from millify import millify

from .....business.transfers import HierarchicalData
from . import IMetricStrategy
from .... import pandas_utils as pd_utl

class StAssetsValues(IMetricStrategy):
    DEFAULT_TOP_PCT = 0.5

    DEFAULT_VALUES_COLS = ['usdValue', 'eurValue', 'ethValue', 'balanceFloat']

    def __init__(self, idx_col: str, cmp_col: str='usdValue', values_cols: str = DEFAULT_VALUES_COLS, top_pct: float = DEFAULT_TOP_PCT):
        self._idx_col = idx_col
        self._cmp_col = cmp_col
        self._values_cols = values_cols
        self._top_pct = top_pct
    
    @property
    def _group_path(self):
        return [['network'], [self._idx_col, 'name'], ['symbol']]

    def clean_df(self, df: pd.DataFrame) -> pd.DataFrame:
        df = df[df[self._cmp_col] > 0].copy()

        df['name'] = df['name'].fillna(df[self._idx_col].str.slice(0,8) + '...')
        df['name'] = df['name'].astype(str)

        return df

    def _build_id(self, dfg, i):
        if i > 0:
            index = [x[0] for x in self._group_path[:i]]
            # numpy.core._exceptions._UFuncNoLoopError: ufunc 'add' did not contain a loop with signature matching types (dtype('<U6'), dtype('float64')) -> None
            return 'total/' + dfg[index].astype(str).agg('/'.join, axis=1)
        else:
            return 'total'

    def _customData(self, dfg):
        if isinstance(dfg, pd.Series):
            return dfg[self._values_cols].apply(millify, precision=2).tolist()
        return pd.Series(dfg[self._values_cols].applymap(millify, precision=2).values.tolist())

    def _flatten(self, lists):
        ret = []

        for i in lists:
            if isinstance(i, list):
                ret.extend(i)
            else:
                ret.append(i)

        return ret

    def _valueBy(self, df, k=1):
        aux = df.groupby(self._flatten(self._group_path[:k]))[self._values_cols].sum()
        return aux

    def _value(self, dfg, i):
        if i < len(self._group_path):
            return 0
        else:
            return dfg[self._cmp_col]

    def process_data(self, df: pd.DataFrame) -> Any:
        df = self.clean_df(df)
        df = df.set_index([self._idx_col, 'network'])

        top, rest = pd_utl.top_rest_daos(df, idx=self._idx_col, value_col=self._cmp_col, top_pct=self._top_pct)

        df_trees = []
        if not top.empty:
            for i, level in enumerate(self._group_path):
                df_tree = pd.DataFrame(columns=['id', 'label', 'parent', 'value', 'customData'])
                dfg = self._valueBy(top, i+1).reset_index()
                
                _id = level[0]
                _label = level[-1]

                df_tree['label'] = dfg[_label].copy().fillna(dfg[_id])
                df_tree['parent'] = self._build_id(dfg, i)
                df_tree['id'] = self._build_id(dfg, i+1)
                df_tree['customData'] = self._customData(dfg)
                df_tree['value'] = self._value(dfg, i+1)

                df_trees.append(df_tree)

        # Now to append the 'Rest' value (symbols parents)
        if not rest.empty:
            df_tree = pd.DataFrame(columns=['id', 'label', 'parent', 'value', 'color', 'customData'])
            dfg = rest.groupby(['network'])[self._values_cols].sum().reset_index()
            df_tree['parent'] = self._build_id(dfg, 1)
            df_tree['id'] = df_tree['parent'] + '/other'
            df_tree['label'] = '<i>Rest of DAOs</i>'
            df_tree['value'] = 0
            df_tree['customData'] = self._customData(dfg)
            df_trees.append(df_tree)

            # Now to append the SYMBOLS into the 'Rest'
            df_tree = pd.DataFrame(columns=['id', 'label', 'parent', 'value', 'color', 'customData'])
            dfg = rest.groupby(['network', 'symbol'])[self._values_cols].sum().reset_index()
            df_tree['parent'] = self._build_id(dfg, 1) + '/other'
            df_tree['id'] = df_tree['parent'] + '/' + dfg['symbol']
            df_tree['label'] = dfg['symbol']
            df_tree['value'] = self._value(dfg, i+1)
            df_tree['customData'] = self._customData(dfg)
            df_trees.append(df_tree)

        df_trees.append(pd.DataFrame({
            'id': ['total'],
            'label': ['All Networks'],
            'parent': [''],
            'value': [0],
            'customData': [self._customData(df[self._values_cols].sum())],
        }))

        df_ret = pd.concat(df_trees, ignore_index=True, axis=0)
        return HierarchicalData.from_df(df_ret)


"""
   Descp: Strategy pattern to create a metric of installed apps.

   Created on: 20-oct-2020

   Copyright 2020-2021 Youssef 'FRYoussef' El Faqir El Rhazoui
        <f.r.youssef@hotmail.com>
"""
import pandas as pd
from typing import List

from dao_analyzer.web.apps.common.data_access.daos.metric.strategy import IMetricStrategy
from dao_analyzer.web.apps.common.business.transfers.stacked_serie import StackedSerie
from dao_analyzer.web.apps.common.business.transfers.serie import Serie
import dao_analyzer.web.apps.common.data_access.pandas_utils as pd_utl


class StInstalledApps(IMetricStrategy):
    __MAX_APPS: int = 15
    __OTHERS: str = 'Others'
    __DF_ID: str = 'id'
    __DF_NAME: str = 'repoName'
    __DF_COUNT: str = 'count'
    __DF_COLS: List[str] = [__DF_ID, __DF_NAME]


    def clean_df(self, df: pd.DataFrame) -> pd.DataFrame:
        dff: pd.DataFrame = df
        dff.loc[:, self.__DF_COLS] = dff[self.__DF_COLS]
        return dff


    def process_data(self, df: pd.DataFrame) -> StackedSerie:
        if pd_utl.is_an_empty_df(df):
            return StackedSerie()
        
        df = self.clean_df(df=df)

        # count pair id-name
        df = pd_utl.count_cols_repetitions(df, self.__DF_COLS, self.__DF_COUNT)

        #remove apps which has not name
        df = df.dropna(subset=[self.__DF_NAME])

        # count name repetitions (should be repoAddress, however not all of them have an entry)
        df = pd_utl.count_cols_repetitions(df, [self.__DF_NAME], self.__DF_COUNT)
        df = df.sort_values(self.__DF_COUNT, ascending=False)

        df = self.__adjust_values(df=df)

        serie: Serie = Serie(x=df[self.__DF_NAME].tolist())
        metric: StackedSerie = StackedSerie(
            serie = serie, 
            y_stack = [df[self.__DF_COUNT].tolist()])

        return metric


    def __adjust_values(self, df: pd.DataFrame) -> pd.DataFrame:
        dff: pd.DataFrame = df
        
        if len(dff) <= self.__MAX_APPS:
            return dff

        names: List[str] = dff[self.__DF_NAME].tolist()[:self.__MAX_APPS]
        installs: List[str] = dff[self.__DF_COUNT].tolist()[:self.__MAX_APPS]
        
        # add the sum of the others
        names.append(self.__OTHERS)
        other_installs: List[str] = dff[self.__DF_COUNT].tolist()[self.__MAX_APPS:]
        installs.append(sum(other_installs))

        return pd.DataFrame(
            {
                self.__DF_NAME: names,
                self.__DF_COUNT: installs
            })




"""
   Descp: Organization list transfer filters.

   Created on: 24-may-2022

   Copyright 2022-2022 David Davó Laviña
        <ddavo@ucm.es>
"""

import abc
from typing import List, Callable, Iterable, Dict, Union
from datetime import datetime, timedelta

from dao_analyzer.web.apps.common.business.transfers.organization.organization import Organization

from dao_analyzer.web.apps.common.resources.strings import TEXT

class Filter(metaclass=abc.ABCMeta):

    def pred(self, org: Organization) -> bool:
        raise NotImplementedError

class OrganizationFilter(Filter, metaclass=abc.ABCMeta):

    def __init__(self, id, title, default):
        self._id = id
        self._title = title
        self._default = default
        self._enabled = default

    @property
    def id(self) -> str:
        return self._id

    @property
    def title(self) -> str:
        return self._title
    
    @property
    def enabled(self) -> bool:
        return self._enabled

    @property
    def default(self) -> bool:
        return self._default

    @enabled.setter
    def enabled(self, value: bool) -> None:
        self._enabled = bool(value)


class ActiveLastYearFilter(OrganizationFilter):
    def __init__(self):
        super().__init__('active-last-year', TEXT['filter_active_last_year_title'], True)

    def pred(self, org: Organization) -> bool:
        """
        Wether the organization has been active in the last year
        """
        la = org.get_last_activity()

        # If la is defined and last activity in the last 365 days
        return la and la >= (datetime.now() - timedelta(days=365))

class NetworkFilter(OrganizationFilter):
    def __init__(self, network: str):
        super().__init__(network, network.capitalize(), True)
        self._network = network

    def pred(self, org: Organization) -> bool:
        """
        Wether the organization is from the selected network
        """
        return org.get_network().lower() == self._network.lower()

class OrganizationFilterGroup(Filter):
    def __init__(self, filters: Iterable[OrganizationFilter] = []):
        self._filters = list(filters)

    def append(self, filter: OrganizationFilter):
        if not isinstance(filter, OrganizationFilter):
            raise TypeError

        self._filters.append(filter)

    def pred(self, org: Organization) -> bool:
        # For an item to be selected, it has to verify every filter
        return all(map(lambda f:f.pred(org), self._filters))

    def get_options(self) -> Dict[str, str]:
        return { x.id:x.title for x in self._filters }
    
    def get_values(self) -> List[str]:
        return [ x.id for x in self._filters if x.enabled ]

class NetworkFilters(OrganizationFilterGroup):
    """ A list of organization filters with an or predicate """

    def __init__(self, networks: List[str]):
        assert(networks) #  Networks can't be null nor empty
        super().__init__(map(NetworkFilter, networks))

    def pred(self, org: Organization) -> bool:
        # An item has to be in any of the selected networks
        return any(map(lambda f:f.pred(org), self._filters))

    def radio_button(self) -> 'NetworkRadioButton':
        return NetworkRadioButton(self)

ALL_NETWORKS_VALUE = '__all_networks'
ALL_NETWORKS_TEXT = TEXT['radio_button_all_networks']

def ALL_NETWORKS_DICT():
    # A new dictionary is needed every time, because they are mutable in Python
    return { ALL_NETWORKS_VALUE : ALL_NETWORKS_TEXT }


class NetworkRadioButton(Filter):
    """ A wrapper around networkFilters to make it compatible with radio buttons """

    def __init__(self, a: Union[NetworkFilters, List[str]], network_value: str = ALL_NETWORKS_VALUE):
        self._options = ALL_NETWORKS_DICT()
        self._value = network_value or ALL_NETWORKS_VALUE

        if isinstance(a, NetworkFilters):
            self._options = a.get_options()
        else:
            self._options = { n:n.capitalize() for n in a }

        if len(self._options) > 1:
            self._options = ALL_NETWORKS_DICT() | dict(sorted(self._options.items(), key=self.sorted_key))
        else:
            # We show the network name, but with the ALL_NETWORKS_VALUE key
            name = next(iter(self._options.values()))
            self._options = { ALL_NETWORKS_VALUE:name }
            self._value = ALL_NETWORKS_VALUE

        if self._value not in self._options.keys():
            raise ValueError('network_value must be in networks')

    @staticmethod
    def sorted_key(item):
        # This function makes mainnet be always the first one
        if item[0] == 'mainnet':
            return '_' + item[0]
        return item[0]

    def get_options(self) -> Dict[str, str]:
        return self._options

    def get_value(self) -> str:
        return self._value

    def pred(self, org: Organization):
        return self._value == ALL_NETWORKS_VALUE or self._value == org.get_network()


SIMPLE_FILTERS: List[Callable[[], OrganizationFilter]] = [
    ActiveLastYearFilter
]



"""
   Descp: Strategy pattern to create a proposal type metric.

   Created on: 9-oct-2020

   Copyright 2020-2021 Youssef 'FRYoussef' El Faqir El Rhazoui
        <f.r.youssef@hotmail.com>
"""
import pandas as pd
from typing import List, Tuple

from dao_analyzer.web.apps.common.data_access.daos.metric.strategy import IMetricStrategy
from dao_analyzer.web.apps.common.business.transfers.stacked_serie import StackedSerie
from dao_analyzer.web.apps.common.business.transfers.serie import Serie
import dao_analyzer.web.apps.common.data_access.pandas_utils as pd_utl


class StProposalType(IMetricStrategy):
    __GRANT_PROPOSAL = 0
    __NEW_MEMBER_PROPOSAL = 1
    __DONATION_PROPOSAL = 2
    __OTHER_PROPOSAL = 3

    __DF_DATE = 'createdAt'
    __DF_COUNT = 'count'
    __DF_TRIBUTE = 'tributeOffered'
    __DF_SHARES = 'sharesRequested'
    __DF_TYPE = 'type'
    __DF_COLS = [__DF_DATE, __DF_TYPE]


    def clean_df(self, df: pd.DataFrame) -> pd.DataFrame:
        dff: pd.DataFrame = df
        dff.loc[:, [self.__DF_DATE, self.__DF_SHARES, self.__DF_TRIBUTE]] = \
            dff[[self.__DF_DATE, self.__DF_SHARES, self.__DF_TRIBUTE]]
        return dff


    def process_data(self, df: pd.DataFrame) -> StackedSerie:
        if pd_utl.is_an_empty_df(df):
            return StackedSerie()

        df = self.clean_df(df=df)
        df = self.__transform_df(df=df)

        df = pd_utl.count_cols_repetitions(df, self.__DF_COLS, self.__DF_COUNT)
        
        # generates a time series
        idx = pd_utl.get_monthly_serie_from_df(df, self.__DF_DATE)
        dff = pd_utl.get_df_from_lists([idx, 0], [self.__DF_DATE, self.__DF_COUNT])
        dff = pd_utl.datetime_to_date(dff, self.__DF_DATE)

        serie: Serie = Serie(x=dff[self.__DF_DATE].tolist())
        grants: List = self.__get_proposals_type(self.__GRANT_PROPOSAL, df, dff)
        new_members: List = self.__get_proposals_type(self.__NEW_MEMBER_PROPOSAL, df, dff)
        donations: List = self.__get_proposals_type(self.__DONATION_PROPOSAL, df, dff)
        others: List = self.__get_proposals_type(self.__OTHER_PROPOSAL, df, dff)

        metric: StackedSerie = StackedSerie(
            serie = serie,
            y_stack = [others, donations, new_members, grants])

        return metric


    def __get_proposals_type(self, typ: int, df: pd.DataFrame, 
    dff: pd.DataFrame) -> List[int]:

        d3f: pd.DataFrame = pd_utl.filter_by_col_value(
            df=df, 
            col=self.__DF_TYPE, 
            value=typ,
            filters=[pd_utl.EQ])

        d3f = d3f.drop(columns=[self.__DF_TYPE])
        d3f = pd.concat([d3f, dff], ignore_index=True)
        d3f = d3f.drop_duplicates(subset=self.__DF_DATE, keep="first")
        d3f = d3f.sort_values(self.__DF_DATE)

        return d3f[self.__DF_COUNT].tolist()


    def __transform_df(self, df: pd.DataFrame) -> pd.DataFrame:
        dff: pd.DataFrame = df
        payments: List[Tuple[int, int]] = \
            tuple(zip(dff[self.__DF_SHARES].tolist(), dff[self.__DF_TRIBUTE].tolist()))
        
        types: List[int] = list(map(self.__get_type, payments))
        dff[self.__DF_TYPE] = types

        # takes just the month
        dff = pd_utl.unix_to_date(dff, self.__DF_DATE)
        dff = pd_utl.transform_to_monthly_date(dff, self.__DF_DATE)

        return dff


    def __get_type(self, tup: Tuple[int, int]) -> int:
        share, tribute = tup
        typ: int = self.__OTHER_PROPOSAL

        if int(share) > 0 and int(tribute) == 0:
            typ = self.__GRANT_PROPOSAL
        elif int(share) > 0 and int(tribute) > 0:
            typ = self.__NEW_MEMBER_PROPOSAL
        elif int(share) == 0 and int(tribute) > 0:
            typ = self.__DONATION_PROPOSAL

        return typ



# AUTO GENERATED FILE - DO NOT EDIT

from dash.development.base_component import Component, _explicitize_args


class PlatformInfo(Component):
    """A PlatformInfo component.
PlatformInfo is the component on the top left of the second header
It receives the DAO name, the network(s) and other data and displays it in a grid view.

Keyword arguments:

- creation_date (string; optional):
    The creation date of the organization.

- name (string; optional):
    The name of the platform.

- networks (list of strings; optional):
    The networks the platform is deployed in.

- participation_stats (list of dicts; optional):
    The array of participation_stats objects."""
    _children_props = []
    _base_nodes = ['children']
    _namespace = 'dao_analyzer_components'
    _type = 'PlatformInfo'
    @_explicitize_args
    def __init__(self, name=Component.UNDEFINED, networks=Component.UNDEFINED, creation_date=Component.UNDEFINED, participation_stats=Component.UNDEFINED, **kwargs):
        self._prop_names = ['creation_date', 'name', 'networks', 'participation_stats']
        self._valid_wildcard_attributes =            []
        self.available_properties = ['creation_date', 'name', 'networks', 'participation_stats']
        self.available_wildcard_properties =            []
        _explicit_args = kwargs.pop('_explicit_args')
        _locals = locals()
        _locals.update(kwargs)  # For wildcard attrs and excess named props
        args = {k: _locals[k] for k in _explicit_args}

        super(PlatformInfo, self).__init__(**args)



"""
   Descp: Strategy pattern to create votes-voters rate.

   Created on: 04-nov-2020

   Copyright 2020-2021 Youssef 'FRYoussef' El Faqir El Rhazoui
        <f.r.youssef@hotmail.com>
"""
import pandas as pd
from typing import List

from dao_analyzer.web.apps.common.data_access.daos.metric.strategy import IMetricStrategy
from dao_analyzer.web.apps.common.business.transfers.stacked_serie import StackedSerie
from dao_analyzer.web.apps.common.business.transfers.serie import Serie 
import dao_analyzer.web.apps.common.data_access.pandas_utils as pd_utl
import dao_analyzer.web.apps.daostack.data_access.daos.metric.strategy.st_total_votes_stakes_option \
    as vote_metric


class StVotesRate(IMetricStrategy):

    VOTES_FOR: int = 0
    VOTES_AGAINST: int = 1


    def __init__(self, m_type) -> None:
        self.__type = m_type


    def clean_df(self, df: pd.DataFrame) -> pd.DataFrame:
        pass


    def process_data(self, df: pd.DataFrame) -> StackedSerie:
        if pd_utl.is_an_empty_df(df):
            return StackedSerie()
        
        votes: StackedSerie = vote_metric\
            .StTotalVSOption(m_type=vote_metric.VOTES).process_data(df)

        rate: List[float] = self.__calculate_rate(m_votes=votes)

        metric: StackedSerie = StackedSerie(
            serie = Serie(x=votes.get_serie()), 
            y_stack = [rate])

        return metric


    def __calculate_rate(self, m_votes: StackedSerie) -> List[float]:
        rates: List[float] = []
        votes_for: List[int] = m_votes.get_i_stack(1)
        votes_against: List[int] = m_votes.get_i_stack(0)
        votes_type: List[int] = votes_against \
            if self.__type == self.VOTES_AGAINST else votes_for

        for i, _ in enumerate(votes_for):
            denominator: int = votes_for[i] + votes_against[i]
            rate: float = votes_type[i] / denominator if denominator else None
            rates.append(rate)

        return rates



"""
   Descp: This file is used as factory to create a DAO metric

   Created on: 5-oct-2020

   Copyright 2020-2021 Youssef 'FRYoussef' El Faqir El Rhazoui
        <f.r.youssef@hotmail.com>
"""
from typing import List, Dict, Tuple
from dao_analyzer.web.apps.common.data_access.daos.metric.strategy import IMetricStrategy

from dao_analyzer.web.apps.common.data_access.daos.metric.metric_dao \
    import MetricDao
from dao_analyzer.web.apps.common.data_access.requesters import CacheRequester, JoinCacheRequester
import dao_analyzer.web.apps.daohaus.data_access.daos.metric.srcs as srcs
from dao_analyzer.web.apps.daohaus.data_access.daos.metric.strategy.st_new_additions import StNewAdditions
from dao_analyzer.web.apps.daohaus.data_access.daos.metric.strategy.st_organization_activity import StOrganizationActivity
from dao_analyzer.web.apps.daohaus.data_access.daos.metric.strategy.st_votes_type import StVotesType
from dao_analyzer.web.apps.daohaus.data_access.daos.metric.strategy.st_active_voters import StActiveVoters
from dao_analyzer.web.apps.daohaus.data_access.daos.metric.strategy.st_proposal_outcome import StProposalOutcome
from dao_analyzer.web.apps.daohaus.data_access.daos.metric.strategy.st_active_members import StActiveMembers
from dao_analyzer.web.apps.daohaus.data_access.daos.metric.strategy.st_proposal_type import StProposalType
from dao_analyzer.web.apps.daohaus.data_access.daos.metric.strategy.st_active_organization import StActiveOrganization
from dao_analyzer.web.apps.daohaus.data_access.daos.metric.strategy.st_approval_proposal_rate import StApprovalProposalRate
from dao_analyzer.web.apps.daohaus.data_access.daos.metric.strategy.st_votes_voters_rate import StVoteVotersRate
from dao_analyzer.web.apps.daohaus.data_access.daos.metric.strategy.st_votes_rate import StVotesRate
from dao_analyzer.web.apps.daohaus.data_access.daos.metric.strategy.st_total_members import StTotalMembers
from dao_analyzer.web.apps.daohaus.data_access.daos.metric.strategy.st_voters_percentage import StVotersPercentage
from dao_analyzer.web.apps.daohaus.data_access.daos.metric.strategy.st_assets_values import StAssetsValues
from dao_analyzer.web.apps.daohaus.data_access.daos.metric.strategy.st_assets_tokens import StAssetsTokens

NEW_MEMBERS = 0
VOTES_TYPE = 1
ACTIVE_VOTERS = 2
OUTGOING_MEMBERS = 3
NEW_PROPOSALS = 4
PROPOSALS_OUTCOME = 5
ACTIVE_MEMBERS = 6
PROPOSAL_TYPE = 7
ACTIVE_ORGANIZATION = 8
APPROVAL_PROPOSAL_RATE = 9
VOTES_VOTERS_RATE = 10
VOTES_FOR_RATE = 11
VOTES_AGAINST_RATE = 12
TOTAL_MEMBERS = 13
VOTERS_PERCENTAGE = 14
ASSETS_VALUES = 15
ASSETS_TOKENS = 16
ORGANIZATION_ACTIVITY = 17

def _metricsDefault(metric: int) -> Tuple[IMetricStrategy, CacheRequester]: # noqa: C901
    requester: CacheRequester = None
    stg = None

    if metric == NEW_MEMBERS:
        stg = StNewAdditions(typ=StNewAdditions.MEMBERS)
        requester = CacheRequester(srcs=[srcs.MEMBERS])
    elif metric == VOTES_TYPE:
        stg = StVotesType()
        requester = CacheRequester(srcs=[srcs.VOTES])
    elif metric == ACTIVE_VOTERS:
        stg = StActiveVoters()
        requester = CacheRequester(srcs=[srcs.VOTES])
    elif metric == OUTGOING_MEMBERS:
        stg = StNewAdditions(typ=StNewAdditions.OUTGOING_MEMBERS)
        requester = CacheRequester(srcs=[srcs.RAGE_QUITS])
    elif metric == NEW_PROPOSALS:
        stg = StNewAdditions(typ=StNewAdditions.PROPOSALS)
        requester = CacheRequester(srcs=[srcs.PROPOSALS])
    elif metric == PROPOSALS_OUTCOME:
        stg = StProposalOutcome()
        requester = CacheRequester(srcs=[srcs.PROPOSALS])
    elif metric == ACTIVE_MEMBERS:
        stg = StActiveMembers()
        requester = CacheRequester(srcs=[
            srcs.PROPOSALS,
            srcs.RAGE_QUITS,
            srcs.VOTES])
    elif metric == PROPOSAL_TYPE:
        stg = StProposalType()
        requester = CacheRequester(srcs=[srcs.PROPOSALS])
    elif metric == ACTIVE_ORGANIZATION:
        stg = StActiveOrganization()
        requester = CacheRequester(srcs=[
            srcs.PROPOSALS,
            srcs.RAGE_QUITS,
            srcs.VOTES])
    elif metric == APPROVAL_PROPOSAL_RATE:
        stg = StApprovalProposalRate()
        requester = CacheRequester(srcs=[srcs.PROPOSALS])
    elif metric == VOTES_VOTERS_RATE:
        stg = StVoteVotersRate()
        requester = CacheRequester(srcs=[srcs.VOTES])
    elif metric == VOTES_FOR_RATE:
        stg = StVotesRate(m_type=StVotesRate.VOTES_FOR)
        requester = CacheRequester(srcs=[srcs.VOTES])
    elif metric == VOTES_AGAINST_RATE:
        stg = StVotesRate(m_type=StVotesRate.VOTES_AGAINST)
        requester = CacheRequester(srcs=[srcs.VOTES])
    elif metric == TOTAL_MEMBERS:
        stg = StTotalMembers()
        requester = CacheRequester(srcs=[
            srcs.MEMBERS,
            srcs.RAGE_QUITS])
    elif metric == VOTERS_PERCENTAGE:
        stg = StVotersPercentage()
        requester = CacheRequester(srcs=[
            srcs.MEMBERS,
            srcs.RAGE_QUITS,
            srcs.VOTES])
    elif metric == ASSETS_VALUES:
        stg = StAssetsValues()
        requester = JoinCacheRequester(srcs=[
            srcs.MOLOCHES,
            srcs.TOKEN_BALANCES
        ])
    elif metric == ASSETS_TOKENS:
        stg = StAssetsTokens()
        requester = CacheRequester(srcs=[
            srcs.MOLOCHES,
            srcs.TOKEN_BALANCES
        ])
    elif metric == ORGANIZATION_ACTIVITY:
        # Same sources as ACTIVE_ORGANIZATIOn
        stg = StOrganizationActivity()
        requester = CacheRequester(srcs=[
            srcs.PROPOSALS,
            srcs.RAGE_QUITS,
            srcs.VOTES
        ])
    else:
        raise ValueError(f"Incorrect metric {metric}")

    return stg, requester

metrics_dict: Dict[int, Tuple[IMetricStrategy, CacheRequester]] = {}

def get_dao(ids: List[str], metric: int) -> MetricDao:
    if metric not in metrics_dict:
        metrics_dict[metric] = _metricsDefault(metric)

    stg, requester = metrics_dict[metric]
    return MetricDao(ids=ids, strategy=stg, requester=requester, address_key='molochAddress')



"""
   Descp: Use this file in order to isolate your app text

   Created on: 21-feb-2020

   Copyright 2020-2021 Youssef 'FRYoussef' El Faqir El Rhazoui
        <f.r.youssef@hotmail.com>
"""

TEXT: dict = {
    'abs_fail': 'Absolute majority rejection',
    'abs_pass': 'Absolute majority approval',
    'active_users_dp_id': 'daostack-dp-activeusers',
    'active_users_dp_title': 'Active holders',
    'active_users_title': 'Active reputation holders',
    'activity_title': 'Activity',
    'approval_proposal_rate_title': 'Approval proposal rate',
    'assets_title': 'Assets',
    'assets_value_dp_id': 'daostack-dp-assets',
    'assets_value_dp_title': 'Treasury',
    'assets_value_title': 'Tokens with fiat equivalence (USD)',
    'assets_novalue_title': 'Tokens without fiat equivalence',
    'boost': 'Rate of boosted proposals that were approved',
    'boost_fail': 'Boosted rejections',
    'boost_pass': 'Boosted approvals',
    'css_id_assets': 'daostack-assets',
    'css_id_activity': 'daostack-activity',
    'css_id_organization': 'daostack-organization',
    'css_id_proposal': 'daostack-proposal',
    'css_id_reputation_holders': 'daostack-reputation-holders',
    'css_id_stake': 'daostack-stake',
    'css_id_votes': 'daostack-votes',
    'css_pane_border': 'daostack-border',
    'different_stakers_title': 'Different stakers',
    'different_voters_title': 'Different voters',
    'disclaimer_activity': 'As activity we consider new proposals, votes and stakes',
    'downstakes': 'Downstakes',
    'new_proposals_dp_id': 'daostack-dp-proposals',
    'new_proposals_dp_title': 'New Proposals',
    'new_proposals_title': 'New proposals',
    'new_users_title': 'New reputation holders',
    'not_boost': 'Rate of non-boosted proposals that were rejected',
    'pane_css_prefix': 'daostack-pane',
    'proposal_boost_outcome_title': 'Closed proposal\'s outcome',
    'proposal_boost_succ_rate_title': 'Success rate of the stakes by type',
    'proposal_outcome_majority_title': 'Closed proposal\'s majority outcome',
    'proposal_title': 'Proposals',
    'proposal_total_succ_rate_title': 'Total success rate of the stakes',
    'queue_fail': 'Non-boosted rejections',
    'queue_pass': 'Non-boosted approvals',
    'rel_fail': 'Relative majority rejection',
    'rep_holder_title': 'Reputation holders',
    'rel_pass': 'Relative majority approval',
    'stake_title': 'Stakes',
    'title_active_organization': 'Number of DAOs active per month',
    'title_organization_activity': 'Total activity per month',
    'total_users_dp_id': 'daostack-dp-users',
    'total_users_dp_title': 'Holders',
    'total_users_title': 'Reputation holders',
    'total_stakes_option_title': 'Total stakes option',
    'total_stakes_title': 'Total stakes',
    'total_votes_option_title': 'Total votes option',
    'total_votes_title': 'Total votes',
    'upstakes': 'Upstakes',
    'vote_against_rate_title': 'Votes-against rate',
    'vote_for_rate_title': 'Votes-for rate',
    'vote_voters_title': 'Number of votes per voter',
    'vote_title': 'Votes',
    'votes_against': 'Votes against',
    'votes_for': 'Votes for',
    'voters_percentage_title': 'Percentage of reputation holders who vote',
}



"""
   Descp: Common elements of the view.

   Created on: 1-apr-2022

   Copyright 2022-2022 David Davó Laviña
        <ddavo@ucm.es>
"""
import pkgutil

from dash import html, dcc
import dash_bootstrap_components as dbc

from ...... import web

ABOUT_MD = pkgutil.get_data(web.__name__, 'assets/ABOUT.md').decode('utf-8')

def get_layout() -> html.Div:
    return dbc.Container(__get_body(), className='top body py-5')

def __get_body() -> html.Div:
    # with open(ABOUT_MD, 'r') as mdf:
    md = dcc.Markdown(ABOUT_MD, className='markdown mt-3')

    back = html.A('Back', href='/', className='about-back')

    return html.Div([back, md], className='col-md-6 mx-auto')


# AUTO GENERATED FILE - DO NOT EDIT

from dash.development.base_component import Component, _explicitize_args


class ParticipationStat(Component):
    """A ParticipationStat component.
This class will be used by DAOInfo and PlatformInfo

Keyword arguments:

- key (string; optional):
    The key used by React for optimization.

- text (string; required):
    The text to show.

- value (string; optional):
    The value to highlight, which will be appended to text."""
    _children_props = []
    _base_nodes = ['children']
    _namespace = 'dao_analyzer_components'
    _type = 'ParticipationStat'
    @_explicitize_args
    def __init__(self, text=Component.REQUIRED, value=Component.UNDEFINED, key=Component.UNDEFINED, **kwargs):
        self._prop_names = ['key', 'text', 'value']
        self._valid_wildcard_attributes =            []
        self.available_properties = ['key', 'text', 'value']
        self.available_wildcard_properties =            []
        _explicit_args = kwargs.pop('_explicit_args')
        _locals = locals()
        _locals.update(kwargs)  # For wildcard attrs and excess named props
        args = {k: _locals[k] for k in _explicit_args}

        for k in ['text']:
            if k not in args:
                raise TypeError(
                    'Required argument `' + k + '` was not specified.')

        super(ParticipationStat, self).__init__(**args)



"""
   Descp: Tester for query builder.

   Created on: 10-mar-2020

   Copyright 2020-2021 Youssef 'FRYoussef' El Faqir El Rhazoui
        <f.r.youssef@hotmail.com>
"""

from typing import List
import unittest
from hypothesis import given, example, settings, strategies as st

from dao_analyzer.web.apps.common.api.graphql.query import Query
from dao_analyzer.web.apps.common.api.graphql.query_builder import QueryBuilder

QueryStrategy = st.builds(
                    Query, 
                    st.text(),
                    st.lists(st.text()),
                    st.dictionaries(
                        keys=st.text(),
                        values=st.text()))


# @st.composite
# def recursive_query(draw) -> Query:
#     return Query(
#         draw(st.text(max_size=10)),
#         draw(QueryStrategy),
#         st.dictionaries(
#             keys=st.text(max_size=10),
#             values=st.text(max_size=10),
#             max_size=10))


class QueryBuilderTest(unittest.TestCase):
    
    @example(queries=None)
    @given(queries=st.lists(QueryStrategy))
    @settings(max_examples=30)
    def test_build_1(self, queries: List[Query]):
        builder: QueryBuilder = QueryBuilder(queries=queries)
        query: str = builder.build()

        # Test that every query has at least curly braces.
        self.assertEqual(f'{query[0]}{query[-1]}', '{}')
        # Test that whether there's at least one query, the string length is
        #  bigger than the first query's header.
        if queries:
            # +2 'cause at least 2 curly braces.
            self.assertGreaterEqual(len(query), len(queries[0].header) + 2)


    def test_build_2(self):
        query1: Query = Query(
                            header='test1',
                            body=['param1', 'param2'],
                            filters={'filt1': 'true', 'filt2': 'false'})
        query2: Query = Query(
                            header='test2',
                            body=Query(
                                header='test3',
                                body=['p1', 'p2'],
                                filters={}
                            ),
                            filters={'f3': True})

        builder: QueryBuilder = QueryBuilder([query1])
        builder.add_query(query2)
        result: str = builder.build()

        sol: str = "{ test1(filt1: true, filt2: false, ){ param1 param2 } \
test2(f3: True, ){ test3(){ p1 p2 } } }"

        self.assertEqual(sol, result, f'Expected = \n{sol}\n found = \n{result}')


if __name__ == "__main__":
    unittest.main()



"""
   Descp: This file is used to store Aragon app's text.

   Created on: 16-oct-2020

   Copyright 2020-2021 Youssef 'FRYoussef' El Faqir El Rhazoui
        <f.r.youssef@hotmail.com>
"""

TEXT: dict = {
    'approved_votes': 'Approved votes',
    'css_pane_border': 'aragon-border',
    'css_id_app': 'aragon-apps',
    'css_id_assets': 'aragon-assets',
    'css_id_cast': 'aragon-casts',
    'css_id_activity': 'aragon-activity',
    'css_id_organization': 'aragon-organization',
    'css_id_token_holder': 'aragon-token-holders',
    'css_id_transactions': 'aragon-transactions',
    'css_id_vote': 'aragon-votes',
    'disclaimer_activity': 'As activity we consider new proposals, votes and transactions',
    'disclaimer_assets': 'Assets from the Recovery Vault of the DAO',
    'dp_id_active_token_holders': 'aragon-dp-token',
    'dp_id_assets': 'aragon-dp-assets',
    'dp_id_new_votes': 'aragon-dp-votes',
    'dp_id_total_token_holders': 'aragon-dp-holders',
    'dp_title_assets': 'Treasury',
    'dp_title_active_token_holders': 'Active Holders',
    'dp_title_new_votes': 'New Proposals',
    'dp_title_total_token_holders': 'Holders',
    'pane_css_prefix': 'aragon-pane',
    'rejected_votes': 'Rejected votes',
    'title_active_organization': 'Number of DAOs active per month',
    'title_active_token_holders': 'Active token holders',
    'title_active_voters': 'Active voters',
    'title_assets': 'Assets',
    'title_assets_value': 'Tokens with fiat equivalence (USD)',
    'title_assets_novalue': 'Tokens without fiat equivalence',
    'title_casted_votes_for_rate': 'Casted votes-for rate',
    'title_casted_votes_against_rate': 'Casted votes-against rate',
    'title_rate_vote_rate': 'Approval vote rate',
    'title_cast_type': 'Casted votes by support',
    'title_installed_apps': 'Installed apps',
    'title_new_transactions': 'New transactions',
    'title_new_votes': 'New votes',
    'title_organization_activity': 'Total activity per month',
    'title_section_activity': 'Activity',
    'title_section_token_holders': 'Token Holders',
    'title_section_vote': 'Votes',
    'title_section_cast': 'Casted votes',
    'title_section_transaction': 'Transactions',
    'title_section_app': 'Aragon apps',
    'title_total_token_holders': 'Total token holders',
    'title_vote_outcome': 'Vote\'s outcome',
    'title_votes_voters': 'Number of casted votes per voter',
    'votes_against': 'Casted votes against',
    'votes_for': 'Casted votes for',
}



"""
   Descp: Interface which must implement all metric strategies.

   Created on: 13-mar-2020

   Copyright 2020-2021 Youssef 'FRYoussef' El Faqir El Rhazoui 
        <f.r.youssef@hotmail.com>
"""

import abc
from typing import Any
from pandas import DataFrame


class IMetricStrategy(metaclass=abc.ABCMeta):
    @classmethod
    def __subclasshook__(cls, subclass):
        return (hasattr(subclass, 'clean_df') and
                callable(subclass.clean_df) and
                hasattr(subclass, 'process_data') and
                callable(subclass.process_data) or
                NotImplemented)


    @abc.abstractmethod
    def clean_df(self, df: DataFrame) -> DataFrame:
        """
        Removes unused columns from df.
        """
        raise NotImplementedError


    @abc.abstractmethod
    def process_data(self, df: DataFrame) -> Any:
        """
        Proces and transform the data frame in a transfer class.
        Return: StackedSerie or NStackedSerie
        """
        raise NotImplementedError



"""
   Descp: Strategy pattern to create approval proposal rate.

   Created on: 05-nov-2020

   Copyright 2020-2021 Youssef 'FRYoussef' El Faqir El Rhazoui
        <f.r.youssef@hotmail.com>
"""
import pandas as pd
from typing import List

from dao_analyzer.web.apps.common.data_access.daos.metric.strategy import IMetricStrategy
from dao_analyzer.web.apps.common.business.transfers.stacked_serie import StackedSerie
from dao_analyzer.web.apps.common.business.transfers.serie import Serie 
import dao_analyzer.web.apps.common.data_access.pandas_utils as pd_utl
import dao_analyzer.web.apps.daohaus.data_access.daos.metric.strategy.st_proposal_outcome \
    as proposal_outcome


class StApprovalProposalRate(IMetricStrategy):

    def clean_df(self, df: pd.DataFrame) -> pd.DataFrame:
        pass


    def process_data(self, df: pd.DataFrame) -> StackedSerie:
        if pd_utl.is_an_empty_df(df):
            return StackedSerie()
        
        outcome: StackedSerie = proposal_outcome.StProposalOutcome()\
            .process_data(df=df)

        rate: List[float] = self.__calculate_rate(outcome)

        metric: StackedSerie = StackedSerie(
            serie = Serie(x=outcome.get_serie()), 
            y_stack = [rate])

        return metric


    def __calculate_rate(self, outcome: StackedSerie) -> List[float]:
        rates: List[float] = []

        approves: List[int] = outcome.get_i_stack(1)
        rejects: List[int] = outcome.get_i_stack(0)

        for i, _ in enumerate(approves):
            numerator: int = approves[i]
            denominator: int = approves[i] + rejects[i]
            rates.append(numerator / denominator if denominator else None)

        return rates



"""
   Descp: A class to warp a query abstraction.

   Created on: 26-feb-2020

   Copyright 2020-2021 Youssef 'FRYoussef' El Faqir El Rhazoui 
        <f.r.youssef@hotmail.com>
"""
from typing import Dict

class Query():
    """
    Atrrs:
        * body: a list of params or a Query object if you want neasted queries.
        * filters: a list of filter in a dictionary.
    """
    
    def __init__(self, header: str = '', body = None, 
    filters: Dict[str, str] = None):

        self.header = header
        # body could be a List[str] or Query 
        self.body = body if body else list()
        self.filters: Dict[str, str] = filters if filters else dict()


"""
    Descp: Daostack Runner and Collectors

    Created on: 15-nov-2021

    Copyright 2021 David Davó
        <david@ddavo.me>
"""
from typing import List, Callable

import pandas as pd
from gql.dsl import DSLField

from .. import config
from ..common.blockscout import BlockscoutBallancesCollector
from ..common.cryptocompare import CCPricesCollector

from ..common import ENDPOINTS, Collector
from ..common.graphql import GraphQLCollector, GraphQLRunner, add_where

def _changeProposalColumnNames(df: pd.DataFrame) -> pd.DataFrame:
    df = df.rename(columns={
        'daoId': 'dao',
        'proposalId': 'proposal'
    })
    return df

def _remove_phantom_daos_wr(daoc: 'DaosCollector') -> Callable[[pd.DataFrame], pd.DataFrame]:
    def _remove_phantom_daos(df: pd.DataFrame) -> pd.DataFrame:
        if df.empty or daoc.df.empty: 
            return df
        
        return df[df.dao.isin(daoc.df.dao)]
    
    return _remove_phantom_daos

class BalancesCollector(BlockscoutBallancesCollector):
    def __init__(self, runner, base, network: str):
        super().__init__(runner, base=base, network=network, addr_key='dao')

class DaosCollector(GraphQLCollector):
    def __init__(self, runner, network: str):
        super().__init__('daos', runner, network=network, endpoint=ENDPOINTS[network]['daostack'])
        
        @self.postprocessor
        def changeColumnNames(df: pd.DataFrame) -> pd.DataFrame:
            df = df.rename(columns={
                'nativeTokenId':'nativeToken',
                'nativeReputationId':'nativeReputation'})
            return df
        
        @self.postprocessor
        def clone_id(df: pd.DataFrame) -> pd.DataFrame:
            if df.empty:
                return df

            df['dao'] = df['id']
            return df

    def query(self, **kwargs) -> DSLField:
        ds = self.schema

        where = { 'register': 'registered' }
        if config.daostack_all:
            where.pop('register')
        
        return ds.Query.daos(**add_where(kwargs, **where)).select(
            ds.DAO.id,
            ds.DAO.name,
            ds.DAO.register,
            ds.DAO.nativeToken.select(ds.Token.id),
            ds.DAO.nativeReputation.select(ds.Rep.id)
        )

class ProposalsCollector(GraphQLCollector):
    def __init__(self, runner, network: str, daoC: DaosCollector):
        super().__init__('proposals', runner, network=network, endpoint=ENDPOINTS[network]['daostack'])

        @self.postprocessor
        def changeColumnNames(df: pd.DataFrame) -> pd.DataFrame:
            return df.rename(columns={
                'daoId': 'dao',
            }).rename(columns=self._stripGenesis)

        @self.postprocessor
        def deleteColums(df: pd.DataFrame) -> pd.DataFrame:
            return df.drop(columns=['competition'], errors='ignore')

        self.postprocessors.append(_remove_phantom_daos_wr(daoC))

    @staticmethod
    def _stripGenesis(s: str):
        tostrip='genesisProtocolParams'

        if s and len(s) > 1 and s.startswith(tostrip):
            s = s[len(tostrip):]
            return s[0].lower() + s[1:]
        else:
            return s

    def query(self, **kwargs) -> DSLField:
        ds = self.schema
        return ds.Query.proposals(**kwargs).select(
            ds.Proposal.id,
            ds.Proposal.proposer,
            # enum ProposalState { None, ExpiredInQueue, Executed, Queued, PreBoosted, Boosted, QuietEndingPeriod}
            ds.Proposal.stage,
            ds.Proposal.createdAt,
            ds.Proposal.preBoostedAt,
            ds.Proposal.boostedAt,
            ds.Proposal.quietEndingPeriodBeganAt,
            ds.Proposal.closingAt,
            ds.Proposal.preBoostedClosingAt,
            ds.Proposal.executedAt,
            ds.Proposal.totalRepWhenExecuted,
            ds.Proposal.totalRepWhenCreated,
            ds.Proposal.executionState,
            ds.Proposal.expiresInQueueAt,
            ds.Proposal.votesFor,
            ds.Proposal.votesAgainst,
            ds.Proposal.winningOutcome,
            ds.Proposal.stakesFor,
            ds.Proposal.stakesAgainst,
            ds.Proposal.genesisProtocolParams.select(
                ds.GenesisProtocolParam.queuedVoteRequiredPercentage,
                ds.GenesisProtocolParam.queuedVotePeriodLimit,
                ds.GenesisProtocolParam.boostedVotePeriodLimit,
                # Used for Holografic Consensus threshold
                ds.GenesisProtocolParam.thresholdConst,
                ds.GenesisProtocolParam.minimumDaoBounty,
                ds.GenesisProtocolParam.daoBountyConst,
            ),
            ds.Proposal.dao.select(ds.DAO.id),
            ds.Proposal.competition.select(ds.CompetitionProposal.id)
        )

class ReputationHoldersCollector(GraphQLCollector):
    def __init__(self, runner, network: str, daoC: DaosCollector):
        super().__init__('reputationHolders', runner, network=network, endpoint=ENDPOINTS[network]['daostack'])
        self.postprocessor(_changeProposalColumnNames)

        self.postprocessors.append(_remove_phantom_daos_wr(daoC))

    def query(self, **kwargs) -> DSLField:
        ds = self.schema
        return ds.Query.reputationHolders(**kwargs).select(
            ds.ReputationHolder.id,
            ds.ReputationHolder.contract,
            ds.ReputationHolder.address,
            ds.ReputationHolder.balance,
            ds.ReputationHolder.createdAt,
            ds.ReputationHolder.dao.select(ds.DAO.id)
        )

class StakesCollector(GraphQLCollector):
    def __init__(self, runner, network: str, daoC: DaosCollector):
        super().__init__('stakes', runner, network=network, endpoint=ENDPOINTS[network]['daostack'])
        self.postprocessor(_changeProposalColumnNames)
    
        self.postprocessors.append(_remove_phantom_daos_wr(daoC))

    def query(self, **kwargs) -> DSLField:
        ds = self.schema
        return ds.Query.proposalStakes(**kwargs).select(
            ds.ProposalStake.id,
            ds.ProposalStake.createdAt,
            ds.ProposalStake.staker,
            ds.ProposalStake.outcome,
            ds.ProposalStake.amount,
            ds.ProposalStake.dao.select(ds.DAO.id),
            ds.ProposalStake.proposal.select(ds.Proposal.id)
        )

class TokenPricesCollector(CCPricesCollector):
    pass

class VotesCollector(GraphQLCollector):
    def __init__(self, runner, network: str, daoC: DaosCollector):
        super().__init__('votes', runner, network=network, endpoint=ENDPOINTS[network]['daostack'])
        self.postprocessor(_changeProposalColumnNames)

        self.postprocessors.append(_remove_phantom_daos_wr(daoC))

    def query(self, **kwargs) -> DSLField:
        ds = self.schema
        return ds.Query.proposalVotes(**kwargs).select(
            ds.ProposalVote.id,
            ds.ProposalVote.createdAt,
            ds.ProposalVote.voter,
            ds.ProposalVote.outcome,
            ds.ProposalVote.reputation,
            ds.ProposalVote.dao.select(ds.DAO.id),
            ds.ProposalVote.proposal.select(ds.Proposal.id)
        )

class CommonRepEventCollector(GraphQLCollector):
    def __init__(self, name, runner, base, network: str): 
        super().__init__(name, runner, network=network, endpoint=ENDPOINTS[network]['daostack'])
        self.base = base

        @self.postprocessor
        def add_dao_id(df: pd.DataFrame) -> pd.DataFrame:
            """ Using the contract info, appends the DAO id.
            Used by ReputationMintsCollector and ReputationBurnsCollector
            """
            # Skip postprocessor if empty
            if df.empty:
                return df
            
            l_index = ['network', 'contract']
            r_index = ['network', 'nativeReputation']

            wants = ['dao']
            prev_cols = list(df.columns)

            # Add the DAO field to the dataframe
            df = df.merge(self.base.df[r_index + wants],
                how='left',
                left_on=l_index,
                right_on=r_index,
            )

            # Get only the dao field
            df = df[prev_cols + wants]
            
            return df

        self.postprocessors.append(_remove_phantom_daos_wr(self.base))

class ReputationMintsCollector(CommonRepEventCollector):
    def __init__(self, *args, **kwargs):
        super().__init__('reputationMints', *args, **kwargs)

    def query(self, **kwargs) -> DSLField:
        ds = self.schema
        return ds.Query.reputationMints(**add_where(kwargs, amount_not=0)).select(
            ds.ReputationMint.id,
            # ds.ReputationMint.txHash, # Not used
            ds.ReputationMint.contract,
            ds.ReputationMint.address,
            ds.ReputationMint.amount,
            ds.ReputationMint.createdAt
        )

class ReputationBurnsCollector(CommonRepEventCollector):
    def __init__(self, *args, **kwargs):
        super().__init__('reputationBurns', *args, **kwargs)

    def query(self, **kwargs) -> DSLField:
        ds = self.schema
        return ds.Query.reputationBurns(**add_where(kwargs, amount_not=0)).select(
            ds.ReputationBurn.id,
            # ds.ReputationBurn.txHash, # Not used
            ds.ReputationBurn.contract,
            ds.ReputationBurn.address,
            ds.ReputationBurn.amount,
            ds.ReputationBurn.createdAt
        )

class DaostackRunner(GraphQLRunner):
    name: str = 'daostack'

    def __init__(self):
        super().__init__()
        self._collectors: List[Collector] = []
        for n in self.networks:
            dc = DaosCollector(self, n)

            self._collectors.extend([
                dc,
                ProposalsCollector(self, n, dc),
                ReputationHoldersCollector(self, n, dc),
                StakesCollector(self, n, dc),
                VotesCollector(self, n, dc),
                BalancesCollector(self, dc, n),
                ReputationMintsCollector(self, dc, n),
                ReputationBurnsCollector(self, dc, n),
            ])

    @property
    def collectors(self) -> List[Collector]:
        return self._collectors



from dao_analyzer.web.apps.common.data_access.requesters.cache_requester import CacheRequester


from typing import List
from pathlib import Path

import pandas as pd

class JoinCacheRequester(CacheRequester):
    """
    A Cache Requester that joins instead of concatenating the dataframes
    """
    
    def __init__(self, srcs: List[Path], on=None, how='outer', dropna=False, columns=None):
        """Initializes the CacheRequester

        Args:
            srcs (List[Path]): A list of paths to read the dfs from
            on: If None, use intersection of columns in all dfs, ignoring the `id`
                column. Defaults to None.
        """
        super().__init__(srcs)
        self._on = on
        self._how = how
        self._dropna = dropna
        self._cols = columns
    
    def tryReload(self):
        dfs = map(pd.read_feather, self._srcs)
        dfs = map(lambda df: df.drop(columns='id'), dfs)

        if self._cols:
            dfs = map(lambda df: df[df.columns.intersection(self._cols)], dfs)

        if self._dropna:
            dfs = map(lambda df: df.dropna(), dfs)

        joined = next(dfs, pd.DataFrame())
        for df in dfs:
            joined = joined.merge(df, how=self._how, on=self._on, suffixes=(None, "_y"))

        self._df = joined


# AUTO GENERATED FILE - DO NOT EDIT

from dash.development.base_component import Component, _explicitize_args


class DAOInfo(Component):
    """A DAOInfo component.
DAOInfo is the component on the top left of the second header
It receives the DAO name, the network(s), the address, the creation date,
and the participation statistics and shows them in a grid view.

Keyword arguments:

- id (string; default 'dao-info'):
    The ID used to identify the component in Dash callbacks.

- address (string; optional):
    The address of the organization.

- creation_date (string; optional):
    The creation date of the organization.

- first_activity (string; optional):
    The date where the first activity was recorded.

- name (string; default 'no name given'):
    The organization (DAO) name.

- network (string; optional):
    The network the DAO is deployed on.

- participation_stats (list of dicts; optional):
    The array of participation_stats objects."""
    _children_props = []
    _base_nodes = ['children']
    _namespace = 'dao_analyzer_components'
    _type = 'DAOInfo'
    @_explicitize_args
    def __init__(self, id=Component.UNDEFINED, name=Component.UNDEFINED, network=Component.UNDEFINED, address=Component.UNDEFINED, creation_date=Component.UNDEFINED, first_activity=Component.UNDEFINED, participation_stats=Component.UNDEFINED, **kwargs):
        self._prop_names = ['id', 'address', 'creation_date', 'first_activity', 'name', 'network', 'participation_stats']
        self._valid_wildcard_attributes =            []
        self.available_properties = ['id', 'address', 'creation_date', 'first_activity', 'name', 'network', 'participation_stats']
        self.available_wildcard_properties =            []
        _explicit_args = kwargs.pop('_explicit_args')
        _locals = locals()
        _locals.update(kwargs)  # For wildcard attrs and excess named props
        args = {k: _locals[k] for k in _explicit_args}

        super(DAOInfo, self).__init__(**args)



"""
   Descp: Strategy pattern to create casted votes-voters rate.

   Created on: 05-nov-2020

   Copyright 2020-2021 Youssef 'FRYoussef' El Faqir El Rhazoui
        <f.r.youssef@hotmail.com>
"""
import pandas as pd
from typing import List

from dao_analyzer.web.apps.common.data_access.daos.metric.strategy import IMetricStrategy
from dao_analyzer.web.apps.common.business.transfers.stacked_serie import StackedSerie
from dao_analyzer.web.apps.common.business.transfers.serie import Serie 
import dao_analyzer.web.apps.common.data_access.pandas_utils as pd_utl
from dao_analyzer.web.apps.aragon.data_access.daos.metric.strategy.st_new_additions \
    import StNewAdditions
from dao_analyzer.web.apps.aragon.data_access.daos.metric.strategy.st_active_voters \
    import StActiveVoters


class StVoteVotersRate(IMetricStrategy):

    def clean_df(self, df: pd.DataFrame) -> pd.DataFrame:
        pass


    def process_data(self, df: pd.DataFrame) -> StackedSerie:
        if pd_utl.is_an_empty_df(df):
            return StackedSerie()
        
        votes: StackedSerie = StNewAdditions(typ=StNewAdditions.VOTE).process_data(df.copy())
        voters: StackedSerie = StActiveVoters().process_data(df)

        rate: List[float] = self.__calculate_rate(
            m_votes=votes,
            m_voters=voters
        )

        metric: StackedSerie = StackedSerie(
            serie = Serie(x=votes.get_serie()), 
            y_stack = [rate])

        return metric


    def __calculate_rate(self, m_votes: StackedSerie, m_voters: StackedSerie) -> List[float]:
        rates: List[float] = []

        votes: List[int] = m_votes.get_i_stack(0)
        voters: List[int] = m_voters.get_i_stack(0)

        for i, _ in enumerate(votes):
            rate: float = round(votes[i] / voters[i], 2) if votes[i] else None
            rates.append(rate)

        return rates



from .hierarchical_data import HierarchicalData
from .n_stacked_serie import NStackedSerie
from .organization import Organization, OrganizationList, Platform
from .serie import Serie
from .stacked_serie import StackedSerie
from .tabular_data import TabularData

__all__ = [
    'HierarchicalData',
    'NStackedSerie',
    'Platform',
    'Organization',
    'OrganizationList',
    'Serie',
    'StackedSerie',
    'TabularData'
]


"""
    Descp: Strategy pattern to add the activity of each DAO per month.
           It'll add proposals created, votes, and stakes.

    Created on: 26-apr-2022

    Copyright 2022-2022 David Davó Laviña
        <ddavo@ucm.es>
"""
import pandas as pd

from dao_analyzer.web.apps.common.business.transfers import Serie, StackedSerie
from dao_analyzer.web.apps.common.data_access.daos.metric.strategy.imetric_strategy import IMetricStrategy
import dao_analyzer.web.apps.common.data_access.pandas_utils as pd_utl

class StOrganizationActivity(IMetricStrategy):
    __DF_DATE = 'createdAt'
    __DF_ID = 'molochAddress'
    __DF_COUNT = 'count'
    __DF_COLS = [__DF_DATE, __DF_ID]

    def clean_df(self, df: pd.DataFrame) -> pd.DataFrame:
        return df[self.__DF_COLS]

    def process_data(self, df: pd.DataFrame) -> StackedSerie:
        if df.empty:
            return StackedSerie()

        df = self.clean_df(df)

        # takes just the month
        df = pd_utl.unix_to_date(df, self.__DF_DATE)
        df = pd_utl.transform_to_monthly_date(df, self.__DF_DATE)

        # join dates-ids
        df = pd_utl.count_cols_repetitions(df, self.__DF_COLS, self.__DF_COUNT)
        # sum of actions per month
        # THIS IS THE ONLY DIFFERENCE WITH StActiveOrganization
        df = pd_utl.sum_cols_repetitions(df, [self.__DF_DATE], self.__DF_COUNT)
        
        # generates a time series
        idx = pd_utl.get_monthly_serie_from_df(df, self.__DF_DATE)
        dff = pd_utl.get_df_from_lists([idx, 0], [self.__DF_DATE, self.__DF_COUNT])
        dff = pd_utl.datetime_to_date(dff, self.__DF_DATE)

        # joinning all the data in a unique dataframe
        df = pd.concat([df, dff], ignore_index=True)
        df = pd_utl.sum_cols_repetitions(df, [self.__DF_DATE], self.__DF_COUNT)
        df = df.sort_values(self.__DF_DATE)

        serie: Serie = Serie(x=df[self.__DF_DATE].tolist())
        metric: StackedSerie = StackedSerie(
            serie = serie, 
            y_stack = [df[self.__DF_COUNT].tolist()])

        return metric



"""
   Descp: Used to wrap the dash instance

   Created on: 20-feb-2020

   Copyright 2020-2021 Youssef 'FRYoussef' El Faqir El Rhazoui 
        <f.r.youssef@hotmail.com>
"""

import argparse
import os
import dash
from dash import dcc, html
import dash_bootstrap_components as dbc

import pandas as pd

from .matomo import Matomo
from .apps.common.resources import colors as COLOR
from .apps.common.resources.strings import TEXT
from .apps.common.presentation.main_view.main_view_controller import bind_callbacks

DEBUG = 'DEBUG' in os.environ and os.environ['DEBUG'].lower() == 'true' or \
        'FLASK_ENV' in os.environ and os.environ['FLASK_ENV'].lower() == 'development'

pd.options.mode.chained_assignment = 'warn' if DEBUG else None

app = dash.Dash(__name__, 
    suppress_callback_exceptions=True, 
    external_scripts=[
    ],
    external_stylesheets=[
        dbc.themes.BOOTSTRAP,
    ],
    meta_tags=[
        { 'name': 'viewport', 'content': 'width=device-width, initial-scale=1' },
    ],
)

if DEBUG:
    pd.options.mode.chained_assignment = None # 'warn' TODO: Delete none
    app.enable_dev_tools(
        debug=True,
    )

matomo_url = None
site_id = None
if 'DAOA_MATOMO_URL' in os.environ:
    matomo_url = os.environ['DAOA_MATOMO_URL']
    site_id = int(os.environ['DAOA_MATOMO_SITE_ID'])

    if matomo_url[-1] != '/': 
        matomo_url += '/'

server = app.server

# see https://dash.plot.ly/external-resources to alter header, footer and favicon
app.title = TEXT['app_title']
index_string = '''
    <!DOCTYPE html>
    <html>
        <head>
            {%metas%}
            <title>{%title%}</title>
            {%favicon%}
            {%css%}
    '''

if matomo_url:
    index_string += Matomo(matomo_url, site_id).get_script()

index_string += '''
        </head>
        <body>
            {%app_entry%}
            {%config%}
            {%scripts%}
            {%renderer%}
        </body>
    </html>
    '''

app.index_string = index_string
app.layout = html.Div([
    dcc.Location(id='url', refresh=False),
    html.Div(id='page-content'),
    dcc.Loading(
        type="cube",
        color=COLOR.DARK_BLUE,
        fullscreen=True,
        children=html.Div(id='header-loading-state', className='d-none')),
])

bind_callbacks(app)

def main():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        '-H', '--host',
        help='Host IP used to serve the application',
        default=os.getenv('HOST', '127.0.0.1'),
    )
    parser.add_argument(
        '-P', '--port',
        help='Port used to serve the application',
        default=os.getenv('PORT', 8050),
    )

    args = parser.parse_args()
    
    app.run_server(
        debug=DEBUG, 
        dev_tools_ui=DEBUG,
        host=args.host,
        port=args.port,
    )



"""
   Descp: This file stores the various sources as an enum

   Created on: 13-dec-2021

   Copyright 2021 David Davó
        <david@ddavo.me>
"""

from pathlib import Path
from typing import List

from dao_analyzer.web.apps.common.data_access.daos.metric.srcs import DATAWAREHOUSE

CACHE_PATH: Path = DATAWAREHOUSE / 'daohaus'

MOLOCHES: Path = CACHE_PATH / 'moloches.arr'
PROPOSALS: Path = CACHE_PATH / 'proposals.arr'
MEMBERS: Path = CACHE_PATH / 'members.arr'
RAGE_QUITS: Path = CACHE_PATH / 'rageQuits.arr'
VOTES: Path = CACHE_PATH / 'votes.arr'
TOKEN_BALANCES: Path = CACHE_PATH / 'tokenBalances.arr'
ALL_FILES: List[str] = [MOLOCHES, PROPOSALS, MEMBERS, RAGE_QUITS, VOTES, TOKEN_BALANCES]


"""
   Descp: This class is used to represent a double scatter figure
          with two sub-figures, one up, the other is below.

   Created on: 14-sep-2020

   Copyright 2020-2021 Youssef 'FRYoussef' El Faqir El Rhazoui
        <f.r.youssef@hotmail.com>
"""
from typing import Dict
import plotly.graph_objs as go
from plotly.subplots import make_subplots

from dao_analyzer.web.apps.common.presentation.charts.layout.figure.figure import Figure
import dao_analyzer.web.apps.common.resources.colors as Color

class DoubleScatterFigure(Figure):

    def __init__(self) -> None:
        super().__init__()
        super().configuration.add_axis(x_axis=2, y_axis=2)
        super().configuration.add_x_params(params={'removemarkers': True}, axis=0)
        super().configuration.add_x_params(params={'tickangle': True}, axis=1)
        super().configuration.add_y_params(params={'reverse_range': True}, axis=1)


    @staticmethod
    def get_empty_plot_data() -> Dict:
        return {
            'common': {
                    'x': list(),
                    'type': '-',
                    'x_format': '',
                    'ordered_keys': [],
                    'y_suffix': ''
        }}


    def get_empty_figure(self) -> Dict:
        return self.get_figure(plot_data = self.get_empty_plot_data())


    def get_figure(self, plot_data: Dict):
        """
        Returns the double scatter chart filled with plot_data.
        Arguments:
            * plot_data = {
                'type_i': { Each scatter type needs an entry in this dict and
                            to be registered in plot_data['common']['ordered_keys'].

                    'x': A list of x-axis values.
                    'y': A list filled with y-axis values.
                    'name': Bar type name.
                    'color': The bar's color.
                    'marker_symbol': Scatter symbol.
                    'position': 'up' or 'down' means the sub-figure to load them.
                }
                'common': { Common values for all the bars
                    'x': A list type filled with values.
                    'type': The type of the x-axis values, e.g. date.
                    'x_format': Format of the x-axis values, e.g. '%b, %Y'.
                    'ordered_keys': An ordered list of the dict-keys of each
                                    bar type. It's used to plot each bar type
                                    in a specific order.
                    'y_suffix': Suffix for the y-axis values.
                }
            } 
        """
        fig = make_subplots(rows=2, cols=1, shared_xaxes=True, vertical_spacing=0)

        for k in plot_data['common']['ordered_keys']:
            fig.add_trace(
                go.Scatter(
                    x=plot_data[k]['x'], 
                    y=plot_data[k]['y'], 
                    marker=dict(
                        color=plot_data[k]['color'],
                        size=6,
                        opacity=0.5,
                        symbol=plot_data[k]['marker_symbol'],
                        line=dict(
                            width=1.2,
                            color=Color.BLACK,
                        )),
                    mode="markers",
                    name=plot_data[k]['name']
                ),
                row=1 if plot_data[k]['position'] == 'up' else 2,
                col=1)

        super().configuration.add_horizontal_line(y=0, y_ref='y')
        super().configuration.add_horizontal_line(y=0, y_ref='y2')
        super().configuration.enable_legend()

        super().configuration.add_x_params(
            params={
                # 'tickvals': plot_data['common']['x'],
                'type': plot_data['common']['type'],
                'tickformat': plot_data['common']['x_format'],
            },
            axis=1)

        super().configuration.add_y_params(
            params={
                'suffix': plot_data['common']['y_suffix'],
                #'matches': 'y2',
            },
            axis=0)

        super().configuration.add_y_params(
            params={
                'suffix': plot_data['common']['y_suffix'],
                #'matches': 'y2',
            },
            axis=1)

        fig.update_layout(
            xaxis=super().configuration.get_x_axis_layout(axis=0),
            xaxis2=super().configuration.get_x_axis_layout(axis=1),
            yaxis=super().configuration.get_y_axis_layout(axis=0),
            yaxis2=super().configuration.get_y_axis_layout(axis=1),
            legend=super().configuration.get_legend(),
            shapes=super().configuration.get_shapes(),
            plot_bgcolor=Color.WHITE
        )

        return fig



"""
   Descp: A factory of dao stacked serie

   Created on: 13-mar-2020

   Copyright 2020-2021 Youssef 'FRYoussef' El Faqir El Rhazoui 
        <f.r.youssef@hotmail.com>
"""
from typing import List, Dict, Tuple
from dao_analyzer.web.apps.common.data_access.daos.metric.strategy import IMetricStrategy

from dao_analyzer.web.apps.common.data_access.daos.metric.metric_dao \
    import MetricDao
from dao_analyzer.web.apps.common.data_access.requesters import CacheRequester, JoinCacheRequester
import dao_analyzer.web.apps.daostack.data_access.daos.metric.srcs as srcs
import dao_analyzer.web.apps.daostack.data_access.daos.metric.strategy.\
    st_time_serie as st_s
import dao_analyzer.web.apps.daostack.data_access.daos.metric.strategy.\
    st_proposal_outcome as st_po
import dao_analyzer.web.apps.daostack.data_access.daos.metric.strategy.\
    st_different_voters_stakers as st_vs
from dao_analyzer.web.apps.daostack.data_access.daos.metric.strategy.\
    st_proposal_majority import StProposalMajority
import dao_analyzer.web.apps.daostack.data_access.daos.metric.strategy.\
    st_total_votes_stakes_option as st_tvso
from dao_analyzer.web.apps.daostack.data_access.daos.metric.strategy.st_active_users\
    import StActiveUsers
from dao_analyzer.web.apps.daostack.data_access.daos.metric.strategy.st_active_organization\
    import StActiveOrganization
from dao_analyzer.web.apps.daostack.data_access.daos.metric.strategy.st_approval_proposal_rate\
    import StApprovalProposalRate
from dao_analyzer.web.apps.daostack.data_access.daos.metric.strategy.st_assets_tokens import StAssetsTokens
from dao_analyzer.web.apps.daostack.data_access.daos.metric.strategy.st_assets_values import StAssetsValues
from dao_analyzer.web.apps.daostack.data_access.daos.metric.strategy.st_votes_voters_rate\
    import StVoteVotersRate
from dao_analyzer.web.apps.daostack.data_access.daos.metric.strategy.st_votes_rate\
    import StVotesRate
from dao_analyzer.web.apps.daostack.data_access.daos.metric.strategy.st_total_reputation_holders\
    import StTotalRepHolders
from dao_analyzer.web.apps.daostack.data_access.daos.metric.strategy.st_voters_percentage\
    import StVotersPercentage
from .strategy.st_organization_activity import StOrganizationActivity

NEW_USERS = 0
NEW_PROPOSALS = 1
PROPOSALS_BOOST_OUTCOME = 2
TOTAL_VOTES = 3
TOTAL_STAKES = 4
DIFFERENT_VOTERS = 5
DIFFERENT_STAKERS = 6
PROPOSAL_MAJORITY = 7
PROPOSALS_TOTAL_SUCCES_RATIO = 8
PROPOSALS_BOOST_SUCCES_RATIO = 9
TOTAL_VOTES_OPTION = 10
TOTAL_STAKES_OPTION = 11
ACTIVE_USERS = 12
ACTIVE_ORGANIZATION = 13
APPROVAL_PROPOSAL_RATE = 14
VOTE_VOTERS_RATE = 15
VOTES_FOR_RATE = 16
VOTES_AGAINST_RATE = 17
TOTAL_REP_HOLDERS = 18
VOTERS_PERCENTAGE = 19
ASSETS_VALUES = 20
ASSETS_TOKENS = 21
ORGANIZATION_ACTIVITY = 22

# TODO: Do this in the other files or MISS
def _metricsDefault(metric: int) -> Tuple[IMetricStrategy, CacheRequester]: # noqa: C901
    requester: CacheRequester = None
    stg = None

    if metric == NEW_USERS:
        stg = st_s.StTimeSerie(st_s.NEW_USERS)
        requester = CacheRequester(srcs=[srcs.REP_HOLDERS])
    elif metric == NEW_PROPOSALS:
        stg = st_s.StTimeSerie(st_s.NEW_PROPOSAL)
        requester = CacheRequester(srcs=[srcs.PROPOSALS])
    elif metric == TOTAL_VOTES:
        stg = st_s.StTimeSerie(st_s.TOTAL_VOTES)
        requester = CacheRequester(srcs=[srcs.VOTES])
    elif metric == TOTAL_STAKES:
        stg = st_s.StTimeSerie(st_s.TOTAL_STAKES)
        requester = CacheRequester(srcs=[srcs.STAKES])
    elif metric == PROPOSALS_BOOST_OUTCOME:
        stg = st_po.StProposalOutcome(st_po.BOOST_OUTCOME)
        requester = CacheRequester(srcs=[srcs.PROPOSALS])
    elif metric == DIFFERENT_VOTERS:
        stg = st_vs.StDifferentVS(st_vs.VOTERS)
        requester = CacheRequester(srcs=[srcs.VOTES])
    elif metric == DIFFERENT_STAKERS:
        stg = st_vs.StDifferentVS(st_vs.STAKERS)
        requester = CacheRequester(srcs=[srcs.STAKES])
    elif metric == PROPOSAL_MAJORITY:
        stg = StProposalMajority()
        requester = CacheRequester(srcs=[srcs.PROPOSALS])
    elif metric == PROPOSALS_TOTAL_SUCCES_RATIO:
        stg = st_po.StProposalOutcome(st_po.TOTAL_SUCCESS_RATIO)
        requester = CacheRequester(srcs=[srcs.PROPOSALS])
    elif metric == PROPOSALS_BOOST_SUCCES_RATIO:
        stg = st_po.StProposalOutcome(st_po.BOOST_SUCCESS_RATIO)
        requester = CacheRequester(srcs=[srcs.PROPOSALS])
    elif metric == TOTAL_VOTES_OPTION:
        stg = st_tvso.StTotalVSOption(st_tvso.VOTES)
        requester = CacheRequester(srcs=[srcs.VOTES])
    elif metric == TOTAL_STAKES_OPTION:
        stg = st_tvso.StTotalVSOption(st_tvso.STAKES)
        requester = CacheRequester(srcs=[srcs.STAKES])
    elif metric == ACTIVE_USERS:
        stg = StActiveUsers()
        requester = CacheRequester(srcs=[
            srcs.PROPOSALS, 
            srcs.VOTES, 
            srcs.STAKES])
    elif metric == ACTIVE_ORGANIZATION:
        stg = StActiveOrganization()
        requester = CacheRequester(srcs=[
            srcs.PROPOSALS,
            srcs.VOTES,
            srcs.STAKES])
    elif metric == APPROVAL_PROPOSAL_RATE:
        stg = StApprovalProposalRate()
        requester = CacheRequester(srcs=[srcs.PROPOSALS])
    elif metric == VOTE_VOTERS_RATE:
        stg = StVoteVotersRate()
        requester = CacheRequester(srcs=[srcs.VOTES])
    elif metric == VOTES_FOR_RATE:
        stg = StVotesRate(m_type=StVotesRate.VOTES_FOR)
        requester = CacheRequester(srcs=[srcs.VOTES])
    elif metric == VOTES_AGAINST_RATE:
        stg = StVotesRate(m_type=StVotesRate.VOTES_AGAINST)
        requester = CacheRequester(srcs=[srcs.VOTES])
    elif metric == TOTAL_REP_HOLDERS:
        stg = StTotalRepHolders()
        requester = CacheRequester(srcs=[srcs.REP_HOLDERS])
    elif metric == VOTERS_PERCENTAGE:
        stg = StVotersPercentage()
        requester = CacheRequester(srcs=[
            srcs.REP_HOLDERS,
            srcs.VOTES])
    elif metric == ASSETS_TOKENS:
        stg = StAssetsTokens()
        requester = JoinCacheRequester(srcs=[
            srcs.DAOS,
            srcs.TOKEN_BALANCES
        ])
    elif metric == ASSETS_VALUES:
        stg = StAssetsValues()
        requester = JoinCacheRequester(srcs=[
            srcs.DAOS,
            srcs.TOKEN_BALANCES
        ])
    elif metric == ORGANIZATION_ACTIVITY:
        stg = StOrganizationActivity()
        requester = CacheRequester(srcs=[
            srcs.PROPOSALS,
            srcs.VOTES,
            srcs.STAKES])
    else:
        raise ValueError(f"Incorrect metric: {metric}")
    
    return stg, requester

metrics_dict: Dict[int, Tuple[IMetricStrategy, CacheRequester]] = {}

def get_dao(ids: List[str], metric: int) -> MetricDao:
    if metric not in metrics_dict:
        metrics_dict[metric] = _metricsDefault(metric)

    stg, requester = metrics_dict[metric]
    return MetricDao(ids=ids, strategy=stg, requester=requester, address_key='dao')



"""
   Descp: Tester for active organization strategy.

   Created on: 23-oct-2020

   Copyright 2020-2021 Youssef 'FRYoussef' El Faqir El Rhazoui
        <f.r.youssef@hotmail.com>
"""

import unittest
from typing import List
import pandas as pd

from test.mocks.unix_date_builder import UnixDateBuilder
from dao_analyzer.web.apps.daostack.data_access.daos.metric.strategy.\
    st_active_organization import StActiveOrganization
from dao_analyzer.web.apps.common.business.transfers.stacked_serie import StackedSerie


class StActiveOrganizationTest(unittest.TestCase):

    def __check_lists(self, df: pd.DataFrame, out: List[int]) -> None:
        strategy: StActiveOrganization = StActiveOrganization()
        result: StackedSerie = strategy.process_data(df=df)

        self.assertListEqual(out, result.get_i_stack(i_stack=0))


    def test_process_data(self):
        bl: UnixDateBuilder = UnixDateBuilder()
        bl.sub(month=2).change(day=25, hour=0, minute=0, second=0)

        in_df: pd.DataFrame = pd.DataFrame([
            {'dao': '0', 'trash': 'trash', 'createdAt': bl.unix()}, #today_year-(today_month-2)-25T00:00:00+00:00
            {'dao': '0', 'trash': 'trash', 'createdAt': bl.unix()}, #today_year-(today_month-2)-25T00:00:00+00:00
            {'dao': '0', 'trash': 'trash', 'createdAt': bl.add(month=1).change(day=1, hour=23).unix()}, #today_year-(today_month-1)-01T23:00:00+00:00
            {'dao': '1', 'trash': 'trash', 'createdAt': bl.sub(month=1).change(minute=59, second=59).unix()}, #today_year-(today_month-2)-01T23:59:59+00:00
            {'dao': '1', 'trash': 'trash', 'createdAt': bl.unix()}, #today_year-(today_month-2)-01T23:59:59+00:00
            {'dao': '3', 'trash': 'trash', 'createdAt': bl.add(month=2).change(day=15, hour=0).unix()}, #today_year-today_month-15T00:00:00+00:00
            {'dao': '2', 'trash': 'trash', 'createdAt': bl.change(day=21).unix()}, #today_year-today_month-21T00:00:00+00:00
            {'dao': '3', 'trash': 'trash', 'createdAt': bl.unix()}, #today_year-today_month-21T00:00:00+00:00
        ])
        out: List[int] = [2, 1, 2]

        self.__check_lists(df=in_df, out=out)


if __name__ == "__main__":
    unittest.main()



"""
   Descp: Tester of percentage of holders who vote.

   Created on: 05-nov-2020

   Copyright 2020-2021 Youssef 'FRYoussef' El Faqir El Rhazoui
        <f.r.youssef@hotmail.com>
"""

import unittest
from typing import List
import pandas as pd

from test.mocks.unix_date_builder import UnixDateBuilder
from dao_analyzer.web.apps.daohaus.data_access.daos.metric.strategy.\
    st_total_members import StTotalMembers 
from dao_analyzer.web.apps.common.business.transfers.stacked_serie import StackedSerie


class StVotersPercentageTest(unittest.TestCase):

    def __check_lists(self, df: pd.DataFrame, out: List[int], stype: int) -> None:
        strategy: StTotalMembers = StTotalMembers()
        result: StackedSerie = strategy.process_data(df=df)

        self.assertListEqual(out, result.get_i_stack(i_stack=0))


    def test_same_size(self):
        bl: UnixDateBuilder = UnixDateBuilder()
        bl.sub(month=2).change(day=25, hour=0, minute=0, second=0)

        in_df: pd.DataFrame = pd.DataFrame([
            {'exists': True, 'createdAt': bl.unix()}, #today_year-(today_month-2)-25T00:00:00+00:00
            {'createdAt': bl.unix()}, #today_year-(today_month-2)-25T00:00:00+00:00
            {'createdAt': bl.add(month=1).change(day=1, hour=23).unix()}, #today_year-(today_month-1)-01T23:00:00+00:00
            {'createdAt': bl.change(second=59, minute=59).unix()}, #today_year-(today_month-1)-01T23:59:00+00:00
            {'exists': True, 'createdAt': bl.add(month=1).unix()}, #today_year-today_month-01T23:59:59+00:00
            {'exists': True, 'createdAt': bl.change(day=21, hour=0).unix()}, #today_year-today_month-21T00:59:00+00:00
        ])
        out: List[int] = [0, -2, 0]

        self.__check_lists(df=in_df, out=out, stype=0)


    def test_news_bigger_size(self):
        bl: UnixDateBuilder = UnixDateBuilder()
        bl.sub(month=2).change(day=25, hour=0, minute=0, second=0)

        in_df: pd.DataFrame = pd.DataFrame([
            {'exists': True, 'createdAt': bl.unix()}, #today_year-(today_month-2)-25T00:00:00+00:00
            {'exists': True, 'createdAt': bl.unix()}, #today_year-(today_month-2)-25T00:00:00+00:00
            {'createdAt': bl.add(month=1).change(day=1, hour=23).unix()}, #today_year-(today_month-1)-01T23:00:00+00:00
            {'exists': True, 'createdAt': bl.change(second=59, minute=59).unix()}, #today_year-(today_month-1)-01T23:59:00+00:00
            {'exists': True, 'createdAt': bl.add(month=1).unix()}, #today_year-today_month-01T23:59:59+00:00
            {'exists': True, 'createdAt': bl.change(day=21, hour=0).unix()}, #today_year-today_month-21T00:59:00+00:00
        ])
        out: List[int] = [2, 2, 4]

        self.__check_lists(df=in_df, out=out, stype=0)


    def test_outs_bigger_size(self):
        bl: UnixDateBuilder = UnixDateBuilder()
        bl.sub(month=2).change(day=25, hour=0, minute=0, second=0)

        in_df: pd.DataFrame = pd.DataFrame([
            {'createdAt': bl.unix()}, #today_year-(today_month-2)-25T00:00:00+00:00
            {'createdAt': bl.unix()}, #today_year-(today_month-2)-25T00:00:00+00:00
            {'createdAt': bl.add(month=1).change(day=1, hour=23).unix()}, #today_year-(today_month-1)-01T23:00:00+00:00
            {'createdAt': bl.change(second=59, minute=59).unix()}, #today_year-(today_month-1)-01T23:59:00+00:00
            {'createdAt': bl.add(month=1).unix()}, #today_year-today_month-01T23:59:59+00:00
            {'createdAt': bl.change(day=21, hour=0).unix()}, #today_year-today_month-21T00:59:00+00:00
        ])
        out: List[int] = [-2, -4, -6]

        self.__check_lists(df=in_df, out=out, stype=0)


if __name__ == "__main__":
    unittest.main()



"""
   Descp: This file is used to store common text for all the apps

   Created on: 1-oct-2020

   Copyright 2020-2021 Youssef 'FRYoussef' El Faqir El Rhazoui
        <f.r.youssef@hotmail.com>
"""

TEXT: dict = {
    'about_us': 'About DAO-Analyzer',
    'acknowledgements': ' and chain-community projects were funded by the European Research Council and The Spanish Ministry of Science and Innovation',
    'active_users_title': 'Active reputation holders',
    'all_orgs': 'All DAOs',
    'app_title': 'DAO-Analyzer',
    'app_title_aragon': 'Aragon-Analyzer',
    'app_title_daohaus': 'DAOhaus-Analyzer',
    'app_title_daostack': 'DAOstack-Analyzer',
    'aragon': 'Aragon',
    'aragon_image_name': 'aragon.png',
    'bt_analyze': 'Analyze',
    'cc_image_name': 'cc-by.png',
    'cc_license_text': 'This site is licensed under a Creative Commons Attribution 4.0 International License',
    'cc_url': 'https://creativecommons.org/licenses/by/4.0/',
    'creation_date_format': '%b %-d, %Y',
    'css_color_app': 'app-color',
    'css_color_aragon': 'aragon-color',
    'css_color_daohaus': 'daohaus-color',
    'css_color_daostack': 'daostack-color',
    'current_version': 'Current version: {version}',
    'dao_analyzer_header_name': 'dao-analyzer_header.svg',
    'dao_analyzer_header_text': 'Explore the DAO world',
    'dao_analyzer_logo_name': 'logo_dao_analyzer.svg',
    'dao_selector_title': 'Select a DAO',
    'daohaus': 'DAOhaus',
    'daohaus_image_name': 'daohaus.png',
    'daostack': 'DAOstack',
    'daostack_image_name': 'daostack.png',
    'default_amount': 'NAN',
    'dp_title_members': 'Members',
    'dp_title_treasury': 'Treasury',
    'ecosystem': 'ecosystem',
    'ecosystem_selector_title': 'Select a platform',
    'erc_image_name': 'logo-erc.png',
    'erc_url': 'https://erc.europa.eu/',
    'filter_active_last_year_title': 'Active last year',
    'follow_us': 'Follow us on ',
    'github': 'GitHub',
    'github_icon_name': 'ico-github.svg',
    'github_url': 'https://github.com/Grasia/dao-analyzer',
    'gpl_image_name': 'logo-GPL.png',
    'gpl_license_text': 'The code is licensed under GPLv3',
    'gpl_url': 'https://www.gnu.org/licenses/gpl-3.0.en.html',
    'graph_amount': '{}: {}',
    'graph_subtitle': '{:.2f}% month over month',
    'grasia_url': 'https://grasia.fdi.ucm.es',
    'no_data': 'No data',
    'no_data_selected': 'There\'s no DAO selected',
    'not_found': 'Error 404: Page not found',
    'last_activity_format': '%b %-d, %Y',
    'last_update': 'Last update:',
    'logo_grasia_name': 'logo-grasia.png',
    'logo_ucm_name': 'logo-ucm.png',
    'network_selector_title': 'Network',
    'p2p_models': 'P2P Models',
    'p2p_models_url': 'https://p2pmodels.eu/',
    'radio_button_all_networks': 'All Networks',
    'spanish_ministry_image_name': 'logo-ministerio.png',
    'spanish_ministry_url': 'https://www.ciencia.gob.es/',
    'stat_members_created_proposals': ' of members have created a proposal',
    'stat_members_created_proposals_novalue': 'No members have created a proposal',
    'stat_members_ever_voted': ' of members have ever voted',
    'stat_members_ever_voted_novalue': 'No members have ever voted',
    'this_month': 'This month',
    'ucm_url': 'https://www.ucm.es/',
    # 'unknown_dao_name': '———', # Em dash
    'unknown_dao_name': 'no name given',
    'url_about': 'about',
    'url_aragon': 'aragon',
    'url_daostack': 'daostack',
    'url_daohaus': 'daohaus',
}



"""
   Descp: This is a dao (data access object) of the organization.
    It's used to drive the data trasformation from datawarehouse to transfer obj.
   Created on: 23-may-2022
   Copyright 2022-2022 David Davó Laviña
        <david@ddavo.me>
"""
import pandas as pd
import numpy as np

from dao_analyzer.web.apps.common.business.transfers import Organization, OrganizationList
from dao_analyzer.web.apps.common.business.transfers.organization.participation_stats import MembersCreatedProposalsStat, MembersEverVotedStat
from dao_analyzer.web.apps.common.business.transfers.organization.platform import Platform
from dao_analyzer.web.apps.common.data_access.daos.platform_dao import PlatformDAO
from dao_analyzer.web.apps.common.data_access.requesters import CacheRequester

from .metric import srcs

class AragonDAO(PlatformDAO):
    __DF_IDX = ['network', 'orgAddress']

    __DF_CAST_DATE = 'createdAt'
    __DF_VOTE_DATE = 'startDate'
    __DF_TRANSACTION_DATE = 'date'
    __DF_MEMBER = 'address'
    __DF_PROPOSER = 'originalCreator'
    __DF_VOTER = 'voter'

    def __init__(self):
        super().__init__(CacheRequester(srcs=[
            srcs.CASTS,
            srcs.VOTES,
            srcs.TRANSACTIONS,
        ]))

        self._orgsCacheRequester = CacheRequester(srcs=[srcs.ORGANIZATIONS])
        self._membersCacheRequester = CacheRequester(srcs=[srcs.TOKEN_HOLDERS])

    def _get_orgs(self) -> pd.DataFrame:
        dforgs: pd.DataFrame = self._orgsCacheRequester.request().set_index(self.__DF_IDX, drop=False)

        # Convert to datetime
        dforgs['createdAt'] = pd.to_datetime(dforgs['createdAt'], unit='s')

        return dforgs

    def _get_members(self) -> pd.DataFrame:
        dfmem: pd.DataFrame = self._membersCacheRequester.request()
        dfmem = dfmem.rename(columns={'organizationAddress': 'orgAddress'})

        members = dfmem[self.__DF_IDX + [self.__DF_MEMBER]]

        return members

    def _get_actions(self) -> pd.DataFrame:
        df = self._requester.request()
        
        # We want to concat all the actions, but they have different column names...
        actions = np.concatenate([
            df[self.__DF_IDX + [self.__DF_CAST_DATE]].dropna().to_numpy(),
            df[self.__DF_IDX + [self.__DF_VOTE_DATE]].dropna().to_numpy(),
            df[self.__DF_IDX + [self.__DF_TRANSACTION_DATE]].dropna().to_numpy(),
        ], axis=0)

        # Back to dataframe (so we can group by)
        actions = pd.DataFrame(actions, columns=self.__DF_IDX + ['actionDate'])
        actions['actionDate'] = pd.to_datetime(actions['actionDate'], unit='s')

        return actions

    def _get_creators(self, members) -> pd.DataFrame:
        creators = self._requester.request()[self.__DF_IDX + [self.__DF_PROPOSER]].dropna()

        # Creators who are also members
        creators = creators.merge(members, 
            left_on = self.__DF_IDX + [self.__DF_PROPOSER],
            right_on = self.__DF_IDX + [self.__DF_MEMBER],
        )[self.__DF_IDX + [self.__DF_PROPOSER]]

        return creators

    def _get_voters(self, members) -> pd.DataFrame:
        voters = self._requester.request()[self.__DF_IDX + [self.__DF_VOTER]].dropna()
        voters[self.__DF_VOTER] = voters[self.__DF_VOTER].str.split('-').str[2]

        # Voters who are also members
        voters = voters.merge(members,
            left_on = self.__DF_IDX + [self.__DF_VOTER],
            right_on = self.__DF_IDX + [self.__DF_MEMBER],
        )[self.__DF_IDX + [self.__DF_VOTER]]

        return voters

    def get_platform(self, orglist: OrganizationList) -> Platform:
        orgs = self._get_orgs()
        members = self._get_members()

        if orglist:
            ids = { (o.get_network(), o.get_id()) for o in orglist }

            orgs = orgs[orgs.index.isin(ids)]
            members = members[members.set_index(self.__DF_IDX).index.isin(orgs.index)]

        mcp_pct = self._get_creators(members)[self.__DF_PROPOSER].nunique() / members[self.__DF_MEMBER].nunique()
        mvp_pct = self._get_voters(members)[self.__DF_VOTER].nunique() / members[self.__DF_MEMBER].nunique()
        creation_date = orgs['createdAt'].min()

        return Platform(
            name = 'Aragon',
            creation_date = creation_date,
            networks = list(orgs['network'].unique()),
            participation_stats = [
                MembersCreatedProposalsStat(mcp_pct),
                MembersEverVotedStat(mvp_pct),
            ]
        )
    
    def get_organization_list(self) -> OrganizationList:
        dforgs = self._get_orgs()
        members = self._get_members()

        # Getting first and last activity
        dfgb = self._get_actions().groupby(self.__DF_IDX)['actionDate']

        dforgs['first_activity'] = dfgb.min()
        dforgs['last_activity'] = dfgb.max()

        # Getting the participation
        gbv = self._get_voters(members).groupby(self.__DF_IDX)
        gbp = self._get_creators(members).groupby(self.__DF_IDX)
        gbm = members.groupby(self.__DF_IDX)

 
        dforgs['mcp_pct'] = gbp[self.__DF_PROPOSER].nunique() / gbm[self.__DF_MEMBER].nunique()
        dforgs['mcv_pct'] = gbv[self.__DF_VOTER].nunique() / gbm[self.__DF_MEMBER].nunique()

        l: OrganizationList = OrganizationList()

        for _, org in dforgs.iterrows():
            l.append(Organization(
                network = org['network'],
                o_id = org['orgAddress'],
                name = org['name'],
                creation_date = org['createdAt'],
                first_activity = self._NaTtoNone(org['first_activity']),
                last_activity = self._NaTtoNone(org['last_activity']),
                participation_stats = [
                    MembersCreatedProposalsStat(org['mcp_pct']),
                    MembersEverVotedStat(org['mcv_pct']),
                ]
            ))

        return l 



"""
   Descp: Tabular data transfer for use in datatables

   Created on: 16-mar-2022

   Copyright 2022-2022 David Davó Laviña
        <ddavo@ucm.es>
"""
from __future__ import annotations
from typing import Dict, List, Any

from dash.dash_table.Format import Format
import pandas as pd

class TabularData():
    def __init__(self):
        self._columns = {}
        self.data = []

    @property
    def columns(self) -> List[Dict[str,str]]:
        return [{'id': id, **rest} for id,rest in self._columns.items()]

    @staticmethod
    def from_df(df: pd.DataFrame) -> TabularData:
        def _t(dtype) -> str:
            if pd.api.types.is_numeric_dtype(dtype):
                return 'numeric'
            return None
        
        df = df.reset_index()
        self = TabularData()
        for id,dtype in df.dtypes.iteritems():
            self._columns[id] = {'name': id}
            if _t(dtype):
                self._columns[id]['type'] = _t(dtype)

        self.data = df.to_dict('records')
        return self

    def _set_column_thing(self, thing, newc: Dict[str, Any]) -> TabularData:
        for id,t in newc.items():
            if id in self._columns:
                self._columns[id][thing] = t
        
        return self
    
    def set_column_names(self, newc: Dict[str, str]) -> TabularData:
        """Sets the column names given the id

        Args:
            newc (Dict[str, str]): A dict of id:'Column Name'
        """
        return self._set_column_thing('name', newc)
    
    def set_column_formats(self, newc: Dict[str, Format]) -> TabularData:
        return self._set_column_thing('format', newc)


"""
   Descp: Controller class for the DataTables

   Created on: 16-mar-2022

   Copyright 2020-2022 David Davó Laviña
        <ddavo@ucm.es>
"""

from dash.dependencies import Input, Output, State
from dao_analyzer.web.apps.common.business.transfers.organization.organization_list import OrganizationList

from dao_analyzer.web.apps.common.business.transfers.tabular_data import TabularData

from .layout import DataTableLayout
from ...business.i_metric_adapter import IMetricAdapter

class DataTableController():
    def __init__(self, table_id: str, layout: DataTableLayout, adapter: IMetricAdapter):
        self.__css_id: str = table_id
        self.__layout: DataTableLayout = layout
        self.__adapter: IMetricAdapter = adapter
    
    @property
    def layout(self) -> DataTableLayout:
        return self.__layout
    
    def bind_callback(self, app) -> None:
        @app.callback(
            [Output(self.__css_id, 'data'), Output(self.__css_id, 'columns')],
            [Input('org-dropdown', 'value')],
            State('org-dropdown', 'options'),
        )
        def update_table(org_id, org_options):
            if not org_id:
                self.__layout.get_layout()

            td: TabularData = self.__adapter.get_plot_data(org_id, OrganizationList.from_dict_representation(org_options))
            return td.data, td.columns


"""
   Descp: Metric adapter common interface.

   Created on: 1-oct-2020

   Copyright 2020-2021 Youssef 'FRYoussef' El Faqir El Rhazoui
        <f.r.youssef@hotmail.com>
"""

import abc
from typing import Dict

class IMetricAdapter(metaclass=abc.ABCMeta):
    @classmethod
    def __subclasshook__(cls, subclass):
        return (hasattr(subclass, 'get_plot_data') and
            callable(subclass.get_plot_data) or
            NotImplemented)


    @abc.abstractmethod
    def get_plot_data(self, dao_id: str, organizations) -> Dict:
        """
        This method it is used to transform business metrics to a dict for 
        its visual representation.
        Arguments: 
                * dao_id: The id of the dao you want to get the data.
        Return:
            A dictionary with the keys to plot it.
        """
        raise NotImplementedError



"""
   Descp: Functions to fetch from endpoint given queries.

   Created on: 10-jul-2020

   Copyright 2020-2021 Youssef 'FRYoussef' El Faqir El Rhazoui 
        <f.r.youssef@hotmail.com>
"""

from gql import Client, gql
from gql.dsl import DSLField, DSLQuery, DSLSchema, DSLType, dsl_gql
from gql.transport.requests import RequestsHTTPTransport
import re
import requests
from functools import partial

import logging
import sys
from tqdm import tqdm
from typing import Dict, List, Union, Iterable

class GQLQueryException(Exception):
    def __init__(self, errors, msg="Errors in GraphQL Query"):
        super().__init__(msg)
        self.__errors = errors

    def errorsString(self) -> str:
        return '\n'.join([f"> {e['message']}" for e in self.__errors])

    def __str__(self):
        return super().__str__() + ":\n" + self.errorsString()

class IndexProgressBar(tqdm):
    def __init__(self, total=0xffff):
        super().__init__(delay=1, total=total, file=sys.stdout, desc="Requesting",
            bar_format="{l_bar}{bar}[{elapsed}<{remaining}{postfix}]", dynamic_ncols=True,
            postfix={"requested":0})
        self.requested = 0

    def progress(self, last_index: str, new_items: int):
        self.requested += new_items
        self.set_postfix(refresh=False, requested=self.requested)

        if not last_index:
            last_index = "0x0"

        match = re.search(r"0x[\da-fA-F]+", last_index)
        if match:
            self.update(int(match[0][:6], 0) - self.n)
        else:
            raise ValueError(f"{last_index} doesn't contain any hex values")

    def complete(self):
        self.update(self.total - self.n)

class RequestProgressSpinner:
    def __init__(self):
        self.prev_lastindex = ""
        self.toFinish = False
        self.total = 0

    def progress(self, last_index: str, new_items: int):
        filler = " " * max(0, len(self.prev_lastindex) - len(last_index))
        self.total += new_items
        print(f"Requesting... Total: {self.total:5d}. Requested until {last_index}"+filler, end='\r', flush=True)
        self.prev_lastindex = last_index
        self.toFinish = True

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        # To remove the last \r
        if self.toFinish:
            print()

    def complete(self):
        pass

class GQLRequester:
    ELEMS_PER_CHUNK: int = 1000

    def __init__(self, endpoint: str, pbar_enabled: bool=True, introspection=True) -> None:
        self.__transport = RequestsHTTPTransport(endpoint)
        self.__client: Client = Client(transport=self.__transport, fetch_schema_from_transport=introspection)
        self.pbar = IndexProgressBar if pbar_enabled else RequestProgressSpinner

        logging.debug(f"Invoked ApiRequester with endpoint: {endpoint}")

    def get_schema(self) -> DSLSchema:
        with self.__client:
            assert(self.__client.schema is not None)
            return DSLSchema(self.__client.schema)

    def request(self, query: DSLQuery | DSLField | str) -> Dict:
        """
        Requests data from endpoint.
        """
        if isinstance(query, DSLField):
            query = DSLQuery(query)

        logging.debug(f"Requesting: {query}")

        if isinstance(query, DSLQuery):
            result = self.__client.execute(dsl_gql(query))
        else:
            result = self.__client.execute(gql(query))
        
        if "errors" in result:
            raise GQLQueryException(result["errors"])

        return result

    def request_single(self, q: DSLQuery | DSLField | str) -> Dict:
        result = self.request(q)
        if result and len(result.values()) == 1:
            return next(iter(result.values()))
        else:
            raise 

    def n_requests(self, query:DSLType, index='id', last_index: str = "", block_hash: str = None) -> List[Dict]:
        """
        Requests all chunks from endpoint.

        Parameters:
            * query: json to request
            * index: dict key to use as index
            * last_index: used to continue the request
            * block_hash: make the request to that block hash
        """
        elements: List[Dict] = list()
        result = Dict
        
        # do-while structure
        exit: bool = False

        with self.pbar() as pbar:
            while not exit:
                query_args = {
                    "where": {index+"_gt": last_index},
                    "first": self.ELEMS_PER_CHUNK
                }
                if block_hash:
                    query_args["block"] = {"hash": block_hash}

                result = self.request_single(query(**query_args))
                elements.extend(result)

                # if return data (result) has no elements, we have finished
                if result: 
                    assert(last_index != result[-1][index])
                    pbar.progress(last_index=last_index, new_items=len(result))
                    last_index = result[-1][index]
                else:
                    pbar.complete()
                    exit = True

        return elements

class CryptoCompareQueryException(Exception):
    def __init__(self, errors, msg="Errors in CryptoCompare Query"):
        super().__init__(msg)
        self.__errors = errors

    def errorsString(self) -> str:
        return self.__errors

    def __str__(self):
        return super().__str__() + ":\n" + self.errorsString()

class CryptoCompareRequester:
    BASEURL = 'https://min-api.cryptocompare.com/data/'

    def __init__(self, api_key: str = None, pbar_enabled: bool = True):
        self.logger = logging.getLogger('ccrequester')
        self.pbar = partial(tqdm, delay=1, file=sys.stdout, desc="Requesting",
            dynamic_ncols=True)
        
        if not api_key:
            logging.warning(f'Invalid api key: {api_key}')
            api_key = ""

        self.api_key = api_key

    def _build_headers(self) -> Dict[str, str]:
        return {
          'Authorization': 'Apikey ' + self.api_key
        }

    def _request(self, url: str, params=None):
        if params is None:
            params = {}

        params['extraParams'] = 'dao-analyzer'

        r = requests.get(url, params=params, headers=self._build_headers())
        self.logger.debug(f'Response status: {r.status_code}, ok: {r.ok}, content: {r.content}')

        # There are two kinds of requests
        # - "Complex" ones which have Response, Message, Type, etc fields
        # - "Simple" ones where the data is the response per se
        if r.ok:
            j = r.json()
            if 'Data' not in j:
                return j
            if "HasWarning" in j and j["HasWarning"]:
                logging.warning("Warning in query", r.url, ":", j["Message"])
            if j["Type"] == 100:
                return j['Data']
        
        raise CryptoCompareQueryException(j["Message"])
    
    def get_available_coin_list(self):
        return self._request(self.BASEURL + 'blockchain/list').values()

    def get_symbols_price(self, fsyms: Union[str, Iterable[str]], tsyms: Iterable[str] = ['USD', 'EUR', 'ETH'], relaxedValidation=False):
        if isinstance(fsyms, str):
            fsyms = [fsyms]
        elif not isinstance(fsyms, Iterable):
            raise TypeError("fsyms must be an Iterable[str] or str")
        else:
            fsyms = list(fsyms)

        mi = 25 # max items
        # Every partition needs to have at least a known value. Else it could fail.
        # That's why we always include BTC
        partitions = [fsyms[i:i+mi]+['BTC'] for i in range(0, len(fsyms), mi)]

        params = {
            'tsyms': ','.join(tsyms),
            'relaxedValidation': str(relaxedValidation).lower()
        }

        ret = {}
        for p in self.pbar(partitions):
            params['fsyms'] = ','.join(p)

            ret.update(self._request(self.BASEURL + 'pricemulti', params=params))

        return ret



from abc import ABC, abstractmethod
from pathlib import Path
from typing import List, Dict, Iterable
import logging
import sys
import json
from datetime import datetime, timezone
import traceback
import pkgutil

from tenacity import retry, retry_if_exception_type, wait_exponential, stop_after_attempt
import pandas as pd
from tqdm import tqdm
from gql.transport.exceptions import TransportQueryError

from .api_requester import GQLRequester
from ..metadata import RunnerMetadata, Block
from .. import config
from ... import cache_scripts

# To be able to obtain endpoints.json
ENDPOINTS: Dict = json.loads(pkgutil.get_data(cache_scripts.__name__, 'endpoints.json'))

def solve_decimals(df: pd.DataFrame) -> pd.DataFrame:
    """ Adds the balanceFloat column to the dataframe

    This column is a precalculated value of tokenBalance / 10 ** tokenDecimals as float
    """
    dkey, bkey, fkey = 'decimals', 'balance', 'balanceFloat'

    df[dkey] = df[dkey].astype(int)
    df[fkey] = df[bkey].astype(float) / 10 ** df[dkey]

    return df

class Collector(ABC):
    INDEX = ['network', 'id']
    
    def __init__(self, name:str, runner):
        self.name: str = name
        self.runner = runner

    @property
    def data_path(self) -> Path:
        return self.runner.basedir / (self.name + '.arr')

    @property
    def long_name(self) -> str:
        return f"{self.runner.name}/{self.name}"

    @property
    def collectorid(self) -> str:
        return self.long_name

    @property
    def df(self) -> pd.DataFrame:
        return pd.DataFrame()

    def verify(self) -> bool:
        """
        Checks if the Collector is in a valid state. This check is run for every
        collector before starting to get data. Can be ignored with --no-verify
        """
        return True

    def _update_data(self, df: pd.DataFrame, force: bool = False) -> pd.DataFrame:
        """ Updates the dataframe in `self.data_path` with the new data.
        """
        if df.empty:
            logging.warning("Empty dataframe, not updating file")
            return

        if not self.data_path.is_file():
            df.reset_index(drop=True).to_feather(self.data_path)
            return

        prev_df: pd.DataFrame = pd.read_feather(self.data_path)

        # If force is selected, we delete the ones of the same network only
        if force:
            prev_df = prev_df[prev_df["network"] != self.network]
        
        prev_df = prev_df.set_index(self.INDEX, verify_integrity=True, drop=True)
        df = df.set_index(self.INDEX, verify_integrity=True, drop=True)

        # Updating data
        combined = df.combine_first(prev_df).reset_index()
        combined.to_feather(self.data_path)
        return combined

    @abstractmethod
    def run(self, force=False, **kwargs) -> None:
        return

class NetworkCollector(Collector):
    """ Collector runnable in a specific network and to a block number """
    def __init__(self, name: str, runner, network: str='mainnet'):
        super().__init__(name, runner) 
        self.network = network

    @property
    def collectorid(self) -> str:
        return '-'.join([super().collectorid, self.network])

class UpdatableCollector(Collector): # Flag class
    pass

class Runner(ABC):
    def __init__(self, dw: Path = Path()):
        self.__dw: Path = dw

    def set_dw(self, dw) -> Path:
        self.__dw = dw

    @property
    def basedir(self) -> Path:
        return self.__dw / self.name

    @property
    def collectors(self) -> List[Collector]:
        return []

    def run(self, **kwargs):
        raise NotImplementedError

class NetworkRunner(Runner, ABC):
    def __init__(self, dw = None):
        super().__init__(dw)
        self.networks = {n for n,v in ENDPOINTS.items() if self.name in v and not n.startswith('_')}

    def filterCollectors(self, 
        networks: Iterable[str] = [],
        names: Iterable[str] = [],
        long_names: Iterable[str] = []
    ) -> Iterable[Collector]:
        result = self.collectors

        if config.only_updatable:
            result = filter(lambda c: isinstance(c, UpdatableCollector), result)

        # networks ^ (names v long_names)
        if networks:
            # GraphQLCollector => c.network in networks
            # a => b : not(a) or b
            result = filter(lambda c: not isinstance(c, NetworkCollector) or c.network in networks, result)

        if names or long_names:
            result = (c for c in result if c.name in names or c.long_name in long_names)

        return result

    def filterCollector(self,
        collector_id: str = None,
        network: str = None,
        name: str = None,
        long_name: str = None,
    ) -> Collector:
        if collector_id:
            return next((c for c in self.collectors if c.collectorid == collector_id), None)

        return next(self.filterCollectors(
            networks=[network] if network else [],
            names=[name] if name else [],
            long_names=[long_name] if long_name else []
        ), None)

    @staticmethod
    @retry(retry=retry_if_exception_type(TransportQueryError), wait=wait_exponential(max=10), stop=stop_after_attempt(3))
    def validated_block(network: str, prev_block: Block = None) -> Block:
        requester = GQLRequester(ENDPOINTS[network]["_blocks"])
        ds = requester.get_schema()

        number_gte = prev_block.number if prev_block else 0
        
        args = {
            "first": 1,
            "skip": config.SKIP_INVALID_BLOCKS,
            "orderBy": "number",
            "orderDirection": "desc",
            "where": {
                "number_gte": number_gte
            }
        }

        if config.block_datetime:
            del args["skip"]
            del args["where"]["number_gte"]
            args["where"]["timestamp_lte"] = int(config.block_datetime.timestamp())

        response = requester.request(ds.Query.blocks(**args).select(
            ds.Block.id,
            ds.Block.number,
            ds.Block.timestamp
        ))["blocks"]
    
        if len(response) == 0:
            logging.warning(f"Blocks query returned no response with args {args}")
            return prev_block

        return Block(response[0])

    @staticmethod
    def _verifyCollectors(tocheck: Iterable[Collector]):
        verified = []
        for c in tqdm(list(tocheck), desc="Verifying"):
            try:
                if c.verify():
                    verified.append(c)
                else:
                    print(f"Verified returned false for {c.collectorid} (view logs the see why)")
            except Exception:
                print(f"Won't run {c.collectorid}", file=sys.stderr)
                traceback.print_exc()
        return verified

    def run(self, networks: List[str] = [], force=False, collectors=None):
        self.basedir.mkdir(parents=True, exist_ok=True)

        print("Verifying collectors")
        verified = self._verifyCollectors(self.filterCollectors(
            networks=networks,
            long_names=collectors
        ))
        if not verified:
            # Nothing to do
            return

        with RunnerMetadata(self) as metadata:
            print(f'--- Updating {self.name} datawarehouse ---')            
            blocks: dict[str, Block] = {}
            for c in verified:
                try:
                    if isinstance(c, NetworkCollector):
                        if c.network not in blocks:
                            # Getting a block more recent than the one in the metadata (just to narrow down the search)
                            print("Requesting a block number...", end='\r')
                            blocks[c.network] = self.validated_block(c.network, None if force else metadata[c.collectorid].block)
                            print(f"Using block {blocks[c.network].id} for {c.network} (ts: {blocks[c.network].timestamp.isoformat()})")

                        print(f"Running collector {c.long_name} ({c.network})")
                        olderBlock = blocks[c.network] < metadata[c.collectorid].block
                        if not force and olderBlock:
                            print("Warning: Forcing because requesting an older block")
                            logging.warning("Forcing because using an older block")

                        # Running the collector
                        c.run(
                            force=force or olderBlock, 
                            block=blocks[c.network],
                            prev_block=metadata[c.collectorid].block,
                        )

                        # Updating the block in the metadata
                        metadata[c.collectorid].block = blocks[c.network]
                    else:
                        print(f"Running collector {c.long_name}")
                        c.run(
                            force=force,
                        )

                    metadata[c.collectorid].last_update = datetime.now(timezone.utc)
                except Exception as e:
                    metadata.errors[c.collectorid] = e.__str__()
                    if config.ignore_errors:
                        print(traceback.format_exc())
                    else:
                        raise e
            print(f'--- {self.name}\'s datawarehouse updated ---')


"""
   Descp: New additions test.

   Created on: 7-oct-2020

   Copyright 2020-2021 Youssef 'FRYoussef' El Faqir El Rhazoui
        <f.r.youssef@hotmail.com>
"""

import unittest
from typing import List
import pandas as pd

from test.mocks.unix_date_builder import UnixDateBuilder
from dao_analyzer.web.apps.daohaus.data_access.daos.metric.strategy.\
    st_new_additions import StNewAdditions

from dao_analyzer.web.apps.common.business.transfers.stacked_serie import StackedSerie


class StNewMembersTest(unittest.TestCase):

    def test_members_process_data(self):
        bl: UnixDateBuilder = UnixDateBuilder()
        bl.sub(month=3).change(day=1, hour=0, minute=0, second=0)

        in_df: pd.DataFrame = pd.DataFrame([
            {'exists': True, 'trash': 'trash', 'createdAt': bl.unix()},#today_year-(today_month-3)-01T00:00:00+00:00
            {'exists': True, 'trash': 'trash', 'createdAt': bl.add(month=1).change(day=25).unix()},#today_year-(today_month-2)-25T00:00:00+00:00
            {'exists': False, 'trash': 'trash', 'createdAt': bl.unix()},#today_year-(today_month-2)-25T00:00:00+00:00
            {'exists': False, 'trash': 'trash', 'createdAt': bl.add(month=1).change(day=1, hour=23).unix()},#today_year-(today_month-1)-01T23:00:00+00:00
            {'exists': True, 'trash': 'trash', 'createdAt': bl.sub(month=1).change(minute=59, second=59).unix()},#today_year-(today_month-2)-01T23:59:59+00:00
            {'exists': True, 'trash': 'trash', 'createdAt': bl.unix()},#today_year-(today_month-2)-01T23:59:59+00:00
            {'exists': True, 'trash': 'trash', 'createdAt': bl.add(month=2, second=1).change(day=21).unix()},#today_year-today_month-21T00:00:00+00:00
        ])
        strategy: StNewAdditions = StNewAdditions(typ=StNewAdditions.MEMBERS)
        result: StackedSerie = strategy.process_data(df=in_df)
        out: List[int] = [1, 3, 0, 1]

        self.assertListEqual(out, result.get_i_stack(i_stack=0))


    def test_outgoing_members_process_data(self):
        bl: UnixDateBuilder = UnixDateBuilder()
        bl.sub(month=3).change(day=4, hour=0, minute=0, second=0)

        in_df: pd.DataFrame = pd.DataFrame([
            {'createdAt': bl.unix()},#today_year-(today_month-3)-04T00:00:00+00:00
            {'createdAt': bl.add(month=1).change(day=16).unix()},#today_year-(today_month-2)-16T00:00:00+00:00
            {'createdAt': bl.unix()},#today_year-(today_month-2)-16T00:00:00+00:00
            {'createdAt': bl.change(day=9, hour=23).unix()},#today_year-(today_month-2)-09T23:00:00+00:00
            {'createdAt': bl.sub(month=1).change(minute=59, second=59).unix()},#today_year-(today_month-3)-01T23:59:59+00:00
        ])
        strategy: StNewAdditions = StNewAdditions(typ=StNewAdditions.OUTGOING_MEMBERS)
        result: StackedSerie = strategy.process_data(df=in_df)
        out: List[int] = [2, 3, 0, 0]

        self.assertListEqual(out, result.get_i_stack(i_stack=0))


if __name__ == "__main__":
    unittest.main()



"""
   Descp: Strategy pattern to create total members.

   Created on: 05-nov-2020

   Copyright 2020-2021 Youssef 'FRYoussef' El Faqir El Rhazoui
        <f.r.youssef@hotmail.com>
"""
import pandas as pd
from typing import List, Tuple

from dao_analyzer.web.apps.common.data_access.daos.metric.strategy import IMetricStrategy
from dao_analyzer.web.apps.common.business.transfers.stacked_serie import StackedSerie
from dao_analyzer.web.apps.common.business.transfers.serie import Serie 
import dao_analyzer.web.apps.common.data_access.pandas_utils as pd_utl
import dao_analyzer.web.apps.daohaus.data_access.daos.metric.strategy.st_new_additions \
    as additions


class StTotalMembers(IMetricStrategy):

    def clean_df(self, df: pd.DataFrame) -> pd.DataFrame:
        pass


    def process_data(self, df: pd.DataFrame) -> StackedSerie:
        if pd_utl.is_an_empty_df(df):
            return StackedSerie()
        
        members_data, rage_quits_data = self.__split_df(df=df)

        new_members: StackedSerie = additions.StNewAdditions(
            typ=additions.StNewAdditions.MEMBERS).process_data(members_data)

        out_members: StackedSerie = additions.StNewAdditions(
            typ=additions.StNewAdditions.OUTGOING_MEMBERS).process_data(rage_quits_data)

        total: List[int] = self.__calculate_total(
            new_members=new_members,
            out_members=out_members)

        s: Serie = Serie(x=new_members.get_serie()
            if len(new_members.get_serie()) >= len(out_members.get_serie())
            else out_members.get_serie())

        metric: StackedSerie = StackedSerie(
            serie = s, 
            y_stack = [total])

        return metric


    def __calculate_total(self, new_members: StackedSerie, out_members: StackedSerie) -> List[int]:
        total: List[int] = []
        
        news, outs = self.__adjust_lists(new_members, out_members)
        total.append(news[0] - outs[0])
        
        for i in range(1, len(news)):
            total.append(news[i] - outs[i] + total[i-1])

        return total


    def __split_df(self, df: pd.DataFrame) -> Tuple[pd.DataFrame, pd.DataFrame]:
        rage_quits_data: pd.DataFrame = df.copy()
        members_data: pd.DataFrame = df.copy() if 'exists' in df.columns else pd.DataFrame() 

        if 'exists' in df.columns:
            # unique attr of member df
            members_data.loc[: ,:] = members_data[members_data.exists.notna()]
            members_data = members_data.dropna(how='all', axis=0)
            rage_quits_data.loc[: ,:] = rage_quits_data[rage_quits_data.exists.isna()]
            rage_quits_data = rage_quits_data.dropna(how='all', axis=0)

        return (members_data, rage_quits_data)


    def __adjust_lists(self, m_news: StackedSerie, 
    m_outs: StackedSerie) -> Tuple[List[int], List[int]]:

        news: List[int] = m_news.get_i_stack(0)
        outs: List[int] = m_outs.get_i_stack(0)

        s1: List = m_news.get_serie()
        s2: List = m_outs.get_serie()

        fill: List[int] = []

        if len(s1) > len(s2):
            fill = self.__fill_holes(s1=s1, s2=s2)
            outs = fill + outs

        elif len(s1) < len(s2):
            fill = self.__fill_holes(s1=s2, s2=s1)
            news = fill + news

        return (news, outs)


    def __fill_holes(self, s1: List, s2: List) -> List[int]:
        fill: List[int] = []

        for s in s1:
            if len(s2) > 0 and s == s2[0]:
                break
            fill.append(0)

        return fill



"""
   Descp: Participation stats DTO.

   Created on: 31-may-2022

   Copyright 2022-2022 David Davó Laviña
        <david@ddavo.me>
"""
from typing import Dict, List, Any

import abc

from dao_analyzer.web.apps.common.resources.strings import TEXT

class ParticipationStat(metaclass=abc.ABCMeta):
    def __init__(self, value):
        self._value = value

    def to_plotly_json(self) -> Dict[str, Any]:
        """ Allows the object to be stored inside a Dcc.Store """
        return {
            'value': self._value,
            'stat': _get_stat_idx(type(self))
        }
    
    @classmethod
    def from_json(cls, d: Dict[str,Any]) -> 'ParticipationStat':
        # 1. Get the class
        statbuilder = STATS_LIST[d['stat']]

        # 2. Instantiate and return
        return statbuilder(d['value'])
    
    @property
    def value(self):
        return self._value

    @property
    def value_str(self):
        if not self.value:
            return None

        if self.value < 0.01:
            return f'{self.value*100:.2g}%'
        return f'{self.value*100:.0f}%'

    @property
    @abc.abstractmethod
    def text(self) -> str:
        raise NotImplementedError

    @classmethod
    @property
    def idx(cls) -> str:
        return str(_get_stat_idx(cls))

    @property
    def key(self) -> str:
        return f'{_get_stat_idx(self)}-{self._value}'

class MembersCreatedProposalsStat(ParticipationStat):

    @property
    def text(self) -> str:
        if self.value is None:
            return TEXT['stat_members_created_proposals_novalue']
        return TEXT['stat_members_created_proposals']

class MembersEverVotedStat(ParticipationStat):

    @property
    def text(self):
        if self.value is None:
            return TEXT['stat_members_ever_voted_novalue']
        return TEXT['stat_members_ever_voted']

STATS_LIST: List = [
    MembersCreatedProposalsStat,
    MembersEverVotedStat,
]

"""
Given an stat class, returns its index in STATS_LIST
"""
def _get_stat_idx(stat):
    return next((i for i,x in enumerate(STATS_LIST) if x == stat), None)


from .DAOInfo import DAOInfo
from .DataPoint import DataPoint
from .ParticipationStat import ParticipationStat
from .PlatformInfo import PlatformInfo

__all__ = [
    "DAOInfo",
    "DataPoint",
    "ParticipationStat",
    "PlatformInfo"
]


#!/usr/bin/env python3
from typing import Dict

from datetime import date
from pathlib import Path
import portalocker as pl
import os
import tempfile
import shutil
from sys import stderr

import logging
from logging.handlers import RotatingFileHandler

from .aragon.runner import AragonRunner
from .daohaus.runner import DaohausRunner
from .daostack.runner import DaostackRunner
from .common import Runner, ENDPOINTS
from .argparser import CacheScriptsArgParser
from . import config

LOG_FILE_FORMAT = "[%(levelname)s] - %(asctime)s - %(name)s - : %(message)s in %(pathname)s:%(lineno)d"
LOG_STREAM_FORMAT = "%(levelname)s: %(message)s"

AVAILABLE_PLATFORMS: Dict[str, Runner] = {
    AragonRunner.name: AragonRunner,
    DaohausRunner.name: DaohausRunner,
    DaostackRunner.name: DaostackRunner
}

# Get available networks from Runners
AVAILABLE_NETWORKS = {n for n in ENDPOINTS.keys() if not n.startswith('_')}

def _call_platform(platform: str, datawarehouse: Path, force: bool=False, networks=None, collectors=None):
    p = AVAILABLE_PLATFORMS[platform]()
    p.set_dw(datawarehouse)
    p.run(networks=networks, force=force, collectors=collectors)

def _is_good_version(datawarehouse: Path) -> bool:
    versionfile = datawarehouse / 'version.txt'
    if not versionfile.is_file():
        return False

    with open(versionfile, 'r') as vf:
        return vf.readline().startswith(config.CACHE_SCRIPTS_VERSION)

def main_aux(datawarehouse: Path):
    if config.delete_force or not _is_good_version(datawarehouse):
        # We skip the dotfiles like .lock
        for p in datawarehouse.glob('[!.]*'):
            if p.is_dir():
                shutil.rmtree(p)
            else:
                p.unlink()

    logger = logging.getLogger()
    logger.propagate = True
    filehandler = RotatingFileHandler(
        filename=config.datawarehouse / 'cache_scripts.log',
        maxBytes=config.LOGGING_MAX_MB * 2**20,
        backupCount=config.LOGGING_BACKUP_COUNT,
    )

    filehandler.setFormatter(logging.Formatter(LOG_FILE_FORMAT))
    logger.addHandler(filehandler)
    logger.setLevel(level=logging.DEBUG if config.debug else logging.INFO)

    # Log errors to STDERR
    streamhandler = logging.StreamHandler(stderr)
    streamhandler.setLevel(logging.WARNING if config.debug else logging.ERROR)
    streamhandler.setFormatter(logging.Formatter(LOG_STREAM_FORMAT))
    logger.addHandler(streamhandler)

    # The default config is every platform
    if not config.platforms:
        config.platforms = AVAILABLE_PLATFORMS.keys()

    # Now calling the platform and deleting if needed
    for p in config.platforms:
        _call_platform(p, datawarehouse, config.force, config.networks, config.collectors)

    # write date
    data_date: str = str(date.today().isoformat())

    if config.block_datetime:
        data_date = config.block_datetime.date().isoformat()

    with open(datawarehouse / 'update_date.txt', 'w') as f:
        f.write(data_date)

    with open(datawarehouse / 'version.txt', 'w') as f:
        f.write(config.CACHE_SCRIPTS_VERSION)

def main_lock(datawarehouse: Path):
    datawarehouse.mkdir(exist_ok=True)
    
    # Lock for the datawarehouse (also used by the dash)
    p_lock: Path = datawarehouse / '.lock'

    # Exclusive lock for the chache-scripts (no two cache-scripts running)
    cs_lock: Path = datawarehouse / '.cs.lock'

    try:
        with pl.Lock(cs_lock, 'w', timeout=1) as lock, \
             tempfile.TemporaryDirectory(prefix="datawarehouse_") as tmp_dw:

            # Writing pid and dir name to lock (debugging)
            tmp_dw = Path(tmp_dw)
            print(os.getpid(), file=lock)
            print(tmp_dw, file=lock)
            lock.flush()

            ignore = shutil.ignore_patterns('*.log', '.lock*')

            # We want to copy the dw, so we open it as readers
            p_lock.touch(exist_ok=True)
            with pl.Lock(p_lock, 'r', timeout=1, flags=pl.LOCK_SH | pl.LOCK_NB):
                shutil.copytree(datawarehouse, tmp_dw, dirs_exist_ok=True, ignore=ignore)

            main_aux(datawarehouse=tmp_dw)

            with pl.Lock(p_lock, 'w', timeout=10):
                shutil.copytree(tmp_dw, datawarehouse, dirs_exist_ok=True, ignore=ignore)

            # Removing pid from lock
            lock.truncate(0)
    except pl.LockException:
        with open(cs_lock, 'r') as f:
            pid = int(f.readline())

        print(f"The cache_scripts are already being run with pid {pid}", file=stderr)
        exit(1)

def main():
    parser = CacheScriptsArgParser(
        available_platforms=list(AVAILABLE_PLATFORMS.keys()),
        available_networks=AVAILABLE_NETWORKS)

    config.populate_args(parser.parse_args())

    if config.display_version:
        print(config.CACHE_SCRIPTS_VERSION)
        exit(0)

    main_lock(config.datawarehouse)

if __name__ == '__main__':
    main()


"""
   Descp: This file stores the various sources as an enum

   Created on: 13-dec-2021

   Copyright 2021 David Davó
        <david@ddavo.me>
"""

from pathlib import Path
from typing import List
from dao_analyzer.web.apps.common.data_access.daos.metric.srcs import DATAWAREHOUSE

CACHE_PATH: Path = DATAWAREHOUSE / 'daostack'

DAOS: Path = CACHE_PATH / 'daos.arr'
PROPOSALS: Path = CACHE_PATH / 'proposals.arr'
REP_HOLDERS: Path = CACHE_PATH / 'reputationHolders.arr'
REP_MINTS: Path = CACHE_PATH / 'reputationMints.arr'
REP_BURNS: Path = CACHE_PATH / 'reputationBurns.arr'
STAKES: Path = CACHE_PATH / 'stakes.arr'
TOKEN_BALANCES: Path = CACHE_PATH / 'tokenBalances.arr'
VOTES: Path = CACHE_PATH / 'votes.arr'

ALL_FILES: List[str] = [
    DAOS, 
    PROPOSALS,
    REP_HOLDERS,
    REP_MINTS,
    REP_BURNS,
    STAKES,
    TOKEN_BALANCES,
    VOTES,
]